<?php
/**
 * @class LSI_Import_Product
 * Handles imports/conversions/syncing of LightSpeed items into WooCommerce.
 */
if ( ! class_exists( 'LSI_Import_Products' ) ) :

    class LSI_Import_Products {

        function __construct() {
            add_action( 'before_delete_post', array( $this, 'sync_deleted_prods' ) );
            add_action( 'wp_ajax_get_prod_ids_ajax', array( $this, 'get_prod_ids_ajax' ) );
            add_action( 'wp_ajax_wclsi_load_prods_ajax', array( $this, 'wclsi_load_prods_ajax' ) );

            add_action( 'wp_ajax_import_all_lightspeed_products_ajax', array(
                $this,
                'import_all_lightspeed_products_ajax'
            ) );
            add_action( 'wp_ajax_import_and_sync_product_ajax', array( $this, 'import_and_sync_product_ajax' ) );
            add_action( 'wp_ajax_import_product_ajax', array( $this, 'import_product_ajax' ) );
            add_action( 'wp_ajax_update_product_ajax', array( $this, 'update_product_ajax' ) );

            add_action( 'wp_ajax_get_import_progress_ajax', array( $this, 'get_import_progress_ajax' ) );
            add_action( 'wp_ajax_manual_prod_update_ajax', array( $this, 'manual_prod_update_ajax' ) );

            add_action( 'wp_ajax_set_prod_sync_ajax', array( $this, 'set_prod_sync_ajax' ) );
            add_action( 'updated_post_meta', array( $this, 'set_sync_value' ), 10, 4 );
            add_action( 'wclsi_create_variation', array( $this, 'create_variation' ), 10, 3 );
            add_action( 'wclsi_update_single_item', array( $this, 'update_single_item' ) );
        }

        /**
         * @param WCLSI_Lightspeed_Prod $wclsi_prod
         * @param false $sync
         * @param false $img_flag
         *
         * @return false|int
         */
        function import_item( WCLSI_Lightspeed_Prod $wclsi_prod, $sync = false, $img_flag = false ) {

            // Used for telling background processes to halt or not run while import is happening
            wp_cache_add( 'wclsi_importing_item', true, WCLSI_CACHE );

            $wclsi_prod = apply_filters( 'wclsi_import_product', $wclsi_prod );
            $post_id    = false;

            try {
                if ( $wclsi_prod->is_matrix_product() ) {
                    set_time_limit( 180 ); // extend to 3 mins for matrix products
                    $post_id = $this->import_matrix_item( $wclsi_prod, $sync, $img_flag );
                } elseif ( $wclsi_prod->is_simple_product() || $wclsi_prod->is_variation_product() ) {
                    $post_id = $this->import_single_item( $wclsi_prod, $sync, $img_flag );
                }
            } catch ( Exception $e ) {
                WCLSI_Logger::log_error( $e->getMessage(), $e, true );
            }

            return $post_id;
        }

        function import_matrix_item( $wclsi_matrix_item, $sync = false, $img_flag = false ) {
            if ( false !== wc_get_product( $wclsi_matrix_item->wc_prod_id ) ) {
                return $wclsi_matrix_item->wc_prod_id;
            }

            global $WCLSI;

            $wclsi_matrix_item = apply_filters( 'wclsi_import_ls_data_matrix_prod', $wclsi_matrix_item );
            $wc_variable_prod  = $this->init_wc_product( $wclsi_matrix_item );
            $this->set_wc_prod_values( $wc_variable_prod, $wclsi_matrix_item );
            $post_id = $wc_variable_prod->save();

            if ( ! is_wp_error( $post_id ) ) {
                // Get & set the category
                $this->handle_product_taxonomy( $wclsi_matrix_item, $post_id );

                $variations = wclsi_get_matrix_prods( $wclsi_matrix_item->item_matrix_id );

                $ls_attr_options = WCLSI_Item_Attributes::get_lightspeed_attribute_options( $variations );
                if ( ! empty( $ls_attr_options ) ) {
                    WCLSI_Item_Attributes::set_product_attributes_for_variable_prod(
                        $wc_variable_prod,
                        $wclsi_matrix_item->item_attribute_set_id,
                        apply_filters( 'wclsi_import_attributes_matrix_item', $ls_attr_options, $post_id )
                    );
                }


                $this->create_variations(
                    apply_filters( 'wclsi_import_variations_matrix_item', $variations, $post_id ),
                    $sync,
                    $post_id
                );


                if ( ! empty( $WCLSI->settings['wclsi_wc_selective_sync']['images'] ) ) {
                    WCLSI_Item_Image::enqueue_item_images_to_download( $wc_variable_prod->get_id(), $wclsi_matrix_item->images );
                }

                if ( $sync ) {
                    update_post_meta( $post_id, WCLSI_SYNC_POST_META, true );
                }

                $this->persist_import_data( $wclsi_matrix_item, $wc_variable_prod, $sync );
            }

            return $post_id;
        }

        function import_single_item( WCLSI_Lightspeed_Prod $ls_prod, $sync, $img_flag = false ) {
            if ( false !== wc_get_product( $ls_prod->wc_prod_id ) ) {
                return $ls_prod->wc_prod_id;
            }

            $ls_prod        = apply_filters( 'wclsi_import_ls_result_single_prod', $ls_prod );
            $wc_simple_prod = $this->init_wc_product( $ls_prod );
            $this->set_wc_prod_values( $wc_simple_prod, $ls_prod );
            $wc_prod_id = $wc_simple_prod->save();

            if ( ! is_wp_error( $wc_prod_id ) ) {
                global $WCLSI;

                $this->handle_product_taxonomy( $ls_prod, $wc_prod_id );

                // Add a sync flag so our cron-job will know to sync this product
                if ( $sync ) {
                    update_post_meta( $wc_prod_id, WCLSI_SYNC_POST_META, true );
                }

                // Get the product images
                if ( ! empty( $WCLSI->settings['wclsi_wc_selective_sync']['images'] ) && ! $img_flag ) {
                    WCLSI_Item_Image::enqueue_item_images_to_download( $wc_prod_id, $ls_prod->images );
                }

                $ls_prod->wc_prod_id  = $wc_prod_id;
                $ls_prod->last_import = current_time( 'mysql' );

                $this->persist_import_data( $ls_prod, $wc_simple_prod, $sync );
            }

            return $wc_prod_id;
        }

        private function log_error_downloading_images( $post_id ) {
            $item = wc_get_product( $post_id );
            $msg  = ( 'There was a timeout downloading images for: "' . $item->get_name() . '". Please try again.' );
            WCLSI_Logger::log_error( $msg, $item, false );

            //delete post to prevent re-importing.
            $this->cleanUpProduct( $post_id );

            return new WP_Error( 'wclsi_bad_import', __( $msg, 'woocomerce-lightspeed-pos' ) );
        }

        private function cleanUpProduct( $post_id ) {
            $product = wc_get_product( $post_id );

            if ( empty( $product ) ) {
                return new WP_Error( 999, sprintf( __( 'No %s is associated with #%d', 'woocommerce' ), 'product', $post_id ) );
            }

            if ( $product->is_type( 'variable' ) ) {
                foreach ( $product->get_children() as $child_id ) {
                    $child = wc_get_product( $child_id );
                    $child->delete( true );
                }
            }

            $product->delete( true );
            $result = $product->get_id() > 0 ? false : true;

            if ( ! $result ) {
                return new WP_Error( 999, sprintf( __( 'This %s cannot be deleted', 'woocommerce' ), 'product' ) );
            }

            // Delete parent product transients.
            if ( $parent_id = wp_get_post_parent_id( $post_id ) ) {
                wc_delete_product_transients( $parent_id );
            }

            return true;
        }

        function update_wc_prod( WCLSI_Lightspeed_Prod $ls_prod, $update_variations = true ) {

            $ls_prod = apply_filters( 'wclsi_update_product', $ls_prod );

            // log manual update
            $is_manual_update = wp_cache_get( 'manual_prod_update', WCLSI_CACHE );
            if ( $is_manual_update ) {

                //unschedule any product update scheduled to go to lightspeed because we are running a manual update
                //to pull inventory from lightspeed
                $get_actions_args     = array(
                    'hook'   => 'wclsi_sync_changes_to_lightspeed',
                    'status' => ActionScheduler_Store::STATUS_PENDING
                );
                $scheduled_syncs      = as_get_scheduled_actions( $get_actions_args, 'OBJECT' );
                $possible_product_ids = array();
                if ( ! empty( $ls_prod->item_id ) ) {
                    $possible_product_ids[] = (int) $ls_prod->wc_prod_id;
                } else {
                    foreach ( $ls_prod->variations as $variation ) {
                        $possible_product_ids[] = (int) $variation->wc_prod_id;
                    }
                }
                foreach ( $scheduled_syncs as $action_id => $action ) {
                    $action_args = $action->get_args();
                    if ( ! empty( $action_args['changes']['id'] ) ) {
                        foreach ( $possible_product_ids as $possible_prod_id ) {
                            if ( $action_args['changes']['id'] === $possible_prod_id ) {
                                as_unschedule_action( 'wclsi_sync_changes_to_lightspeed', $action_args );
                                WCLSI_Logger::log_debug_event( "Cancelled sync for Woo Prod ID $possible_prod_id because manual update was run" );
                            }
                        }

                    }
                }
                $msg  = "-------------- Manual Update --------------";
                $data = array(
                    "description"    => $ls_prod->description,
                    "wc product id " => $ls_prod->wc_prod_id
                );
                WCLSI_Logger::log_sync_event( $msg, $data );
            }

            if ( $ls_prod->is_simple_product() || $ls_prod->is_variation_product() ) {
                return $this->update_single_item( $ls_prod );
            } elseif ( $ls_prod->is_matrix_product() ) {
                return $this->update_matrix_item( $ls_prod, $update_variations );
            }

            return new WP_Error( 'wclsi_bad_update', __( 'Could not process update, invalid product!', 'woocomerce-lightspeed-pos' ) );
        }

        function update_matrix_variations( $variations, WC_Product_Variable $wc_parent_product ) {
            $new_variations_to_import = array();

            // Get the itemAttributeSetID from a random variation
            $attr_set_id = $variations[0]->item_attributes->itemAttributeSetID;
            if ( $attr_set_id > 0 ) {
                // Re-set attributes in case new variation introduces new attribute(s)
                $wclsi_parent_prod = new WCLSI_Lightspeed_Prod();
                $wclsi_parent_prod->init_via_wc_prod_id( $wc_parent_product->get_id() );
                $all_variations = array_merge( $wclsi_parent_prod->variations, $variations );

                global $WCLSI;
                $selective_sync_settings = $WCLSI->settings[ WCLSI_WC_SELECTIVE_SYNC ];
                if ( isset( $selective_sync_settings['attributes'] ) && 'true' === $selective_sync_settings['attributes'] ) {
                    $ls_attr_options = WCLSI_Item_Attributes::get_lightspeed_attribute_options( $all_variations );
                    if ( ! empty( $ls_attr_options ) ) {
                        WCLSI_Item_Attributes::set_product_attributes_for_variable_prod(
                            $wc_parent_product,
                            $attr_set_id,
                            apply_filters(
                                'wclsi_update_attributes_matrix_item',
                                $ls_attr_options,
                                $wc_parent_product->get_id()
                            )
                        );
                    }
                }

                /**
                 * Update attributes for all variations just in case we are adding a new product or
                 * the call to "set_item_attributes()" has caused a conversion of a custom attribute to a
                 * taxonomy one.
                 */
                foreach ( $all_variations as $wclsi_variation ) {
                    if ( $wclsi_variation->wc_prod_id > 0 ) {
                        $wc_variation_prod = wc_get_product( $wclsi_variation->wc_prod_id );
                        if ( $wclsi_variation->item_matrix_id > 0 ) {
                            WCLSI_Item_Attributes::set_attributes_for_wc_variation( $wc_variation_prod, $wclsi_variation );
                        } else {
                            WCLSI_Logger::log_error( "Error importing variation. Product is not of variation type: ", $wclsi_variation->description );
                        }
                    }
                }
            }

            foreach ( $variations as $variation ) {
                if ( $variation->wc_prod_id > 0 ) {
                    if ( count( $variations ) > WCLSI_VARIATION_LIMIT ) {
                        as_schedule_single_action( time(), 'wclsi_update_single_item', array( $variation->id ), 'wclsi' );
                    } else {
                        $this->update_single_item( $variation );
                    }
                } elseif ( is_null( $variation->wc_prod_id ) ) {
                    $new_variations_to_import[] = $variation;
                }
            }

            if ( ! empty( $new_variations_to_import ) ) {
                $sync = get_post_meta( $wc_parent_product->get_id(), WCLSI_SYNC_POST_META, true );
                $this->create_variations( $new_variations_to_import, $sync, $wc_parent_product->get_id() );
            }
        }


        /**
         * Uses the 'delete_post' action hook when a product gets deleted and removes the "last_import" and
         * "last_sync" fields of the product in the wclsi cache.
         *
         * @param $wc_prod_id
         */
        function sync_deleted_prods( $wc_prod_id ) {
            $ls_prod = new WCLSI_Lightspeed_Prod();
            $ls_prod->init_via_wc_prod_id( $wc_prod_id );
            $ls_prod->clear_wc_prod_association();
        }

        function get_prod_ids_ajax() {
            wclsi_verify_nonce();

            if ( isset( $_POST['init_import'] ) && true == $_POST['init_import'] ) {
                $prod_ids = wclsi_get_ls_prod_ids();
                echo json_encode( $prod_ids );
                exit;
            }
        }

        function import_all_lightspeed_products_ajax() {
            //	$sync = ( isset( $_POST[ 'sync_flag' ] ) && $_POST[ 'sync_flag' ] == 'true' ) ? true : false;
            $this->process_product_ajax_request( 'import' );
        }

        function import_product_ajax() {
            $this->process_product_ajax_request( 'import', false );
        }

        function import_and_sync_product_ajax() {
            $this->process_product_ajax_request( 'import_and_sync' );
        }

        function update_product_ajax() {
            $this->process_product_ajax_request( 'update' );
        }

        /**
         * Imports and adds a sync flag to all the loaded products from LightSpeed
         *
         * @param - $_POST['init_import'] - Will return all existing prods
         * @param bool $sync
         */
        function process_product_ajax_request( $action, $sync = true ) {
            wclsi_verify_nonce();

            $prod_id = isset( $_POST['prod_id'] ) ? (int) $_POST['prod_id'] : false;

            if ( false !== $prod_id ) {
                $ls_item = new WCLSI_Lightspeed_Prod( $prod_id );

                if ( $ls_item->id > 0 ) {
                    if ( $action == 'import' || $action == 'import_and_sync' ) {
                        $result = $this->import_item( $ls_item, $sync );
                    } else if ( $action == 'update' ) {
                        wp_cache_add( 'manual_prod_update', true, WCLSI_CACHE );
                        $result = $ls_item->update_via_api();
                    }

                    if ( ! is_wp_error( $result ) && $action === 'update' && $ls_item->wc_prod_id > 0 ) {
                        $ls_item->reload();
                        $result = $this->update_wc_prod( $ls_item, true );
                    }

                    if ( ! is_wp_error( $result ) ) {
                        echo json_encode(
                            array(
                                'success'   => $result,
                                'prod_id'   => $prod_id,
                                'WAIT_TIME' => wclsi_get_api_wait_time_ms()
                            )
                        );
                    } else {
                        echo json_encode(
                            array(
                                'errors' => array(
                                    array( 'message' => $result->get_error_message() )
                                )
                            )
                        );
                        exit;
                    }
                } else {
                    echo json_encode(
                        array(
                            'errors' => array(
                                array( 'message' => sprintf( __( 'Product with ID %d does not exist!', 'woocommerce-lightspeed-pos' ), $prod_id ) )
                            )
                        )
                    );
                    exit;
                }
            } else {
                echo json_encode(
                    array(
                        'errors' => array(
                            array( 'message' => __( 'Could not find a product ID!', 'woocommerce-lightspeed-pos' ) )
                        )
                    )
                );
                exit;
            }
            exit;
        }

        /**
         * Starts the load process via an AJAX call. Returns the total number of products to the JS script.
         * JS script then makes (at most) 40 calls a minute as to not exceed API throttling.
         */
        function wclsi_load_prods_ajax() {
            wclsi_verify_nonce();

            $offset = 0;
            if ( isset( $_POST['offset'] ) ) {
                $offset = (int) $_POST['offset'];
            }

            $limit = 1;
            if ( isset( $_POST['limit'] ) ) {
                $limit = (int) $_POST['limit'];
            }

            // Flag to get total count of products to load
            $get_count = $_POST['getCount'] === 'true' ? true : false;

            $load_matrix_prods = $_POST['getMatrix'] == 'true' ? true : false;

            if ( $load_matrix_prods ) {
                $result          = $this->pull_matrix_items_by_offset( $offset, $limit );
                $prod_identifier = 'ItemMatrix';
            } else {
                $result          = $this->pull_simple_items_by_offset( $offset, $limit );
                $prod_identifier = 'Item';
            }

            if ( isset( $result ) && ! is_wp_error( $result ) ) {
                if ( $get_count ) {
                    $count = ceil( $result->{'@attributes'}->count / 100 ) * 100;

                    echo json_encode(
                        array(
                            'count'      => $count,
                            'real_count' => $result->{'@attributes'}->count,
                            'errors'     => get_settings_errors( 'wclsi_settings' ),
                            'WAIT_TIME'  => wclsi_get_api_wait_time_ms()
                        )
                    );
                } else {
                    if ( isset( $result->{$prod_identifier} ) ) {
                        $ls_api_items = $result->{$prod_identifier};
                    } else {
                        $ls_api_items = null;
                    }

                    if ( isset( $ls_api_items ) ) {
                        if ( is_object( $ls_api_items ) ) {
                            $single_item  = $ls_api_items;
                            $ls_api_items = array( $single_item );
                        }

                        foreach ( $ls_api_items as $item ) {
                            WCLSI_Lightspeed_Prod::insert_ls_api_item( $item );
                        }
                    }

                    if ( ( $offset + $limit ) >= $result->{'@attributes'}->count ) { // Last chunk condition
                        global $WCLSI;
                        date_default_timezone_set( $WCLSI->store_timezone );
                        update_option( WCLSI_LAST_LOAD_TIMESTAMP, date( DATE_ATOM ) );

                        // We load matrix prods after simple prods, before we finish we should pull the item_attr_sets
                        if ( $load_matrix_prods ) {
                            self::pull_item_attribute_sets();
                        }

                        echo json_encode(
                            array(
                                'loaded_chunk_' . $offset => true,
                                'redirect'                => WCLSI_ADMIN_URL,
                                'errors'                  => get_settings_errors( 'wclsi_settings' ),
                                'matrix'                  => $load_matrix_prods,
                                'WAIT_TIME'               => wclsi_get_api_wait_time_ms()
                            )
                        );

                    } else {
                        echo json_encode(
                            array(
                                'loaded_chunk_' . $offset => true,
                                'errors'                  => get_settings_errors( 'wclsi_settings' ),
                                'matrix'                  => $load_matrix_prods,
                                'WAIT_TIME'               => wclsi_get_api_wait_time_ms()
                            )
                        );
                    }
                }

            } elseif ( is_wp_error( $result ) ) {
                echo json_encode(
                    array(
                        'loaded_chunk_' . $offset => true,
                        'errors'                  => $result->get_error_message(),
                        'matrix'                  => $load_matrix_prods,
                        'WAIT_TIME'               => wclsi_get_api_wait_time_ms()
                    )
                );
            }
            exit;
        }

        function pull_simple_items_by_offset( $offset = 0, $limit = 100 ) {
            global $WCLSI, $WCLSI_SINGLE_LOAD_RELATIONS;

            $search_params = array(
                'load_relations' => json_encode( $WCLSI_SINGLE_LOAD_RELATIONS ),
                'offset'         => $offset,
                'limit'          => $limit,
                'archived'       => $WCLSI->settings[ WCLSI_IGNORE_ARCHIVED_LS_PRODS ] ? 'false' : 'true'
            );

            $search_params = apply_filters( 'wclsi_ls_import_prod_params', $search_params );

            return $WCLSI->make_api_call( 'Account/' . $WCLSI->ls_account_id . '/Item/', 'Read', $search_params );
        }

        function pull_matrix_items_by_offset( $offset = 0, $limit = 100 ) {
            global $WCLSI, $WCLSI_MATRIX_LOAD_RELATIONS;

            $resource      = 'Account/' . $WCLSI->ls_account_id . '/ItemMatrix';
            $search_params = array(
                'offset'         => $offset,
                'load_relations' => json_encode( $WCLSI_MATRIX_LOAD_RELATIONS ),
                'limit'          => $limit,
                'archived'       => (bool) $WCLSI->settings[ WCLSI_IGNORE_ARCHIVED_LS_PRODS ] ? 'false' : 'true'
            );

            $search_params = apply_filters( 'wclsi_ls_import_matrix_params', $search_params );

            return $WCLSI->make_api_call( $resource, 'Read', $search_params );
        }

        function pull_item_attribute_sets() {
            global $WCLSI;
            $item_attr_sets = $WCLSI->make_api_call(
                "Account/{$WCLSI->ls_account_id}/ItemAttributeSet",
                "Read"
            );

            if ( is_object( $item_attr_sets ) && property_exists( $item_attr_sets, 'ItemAttributeSet' ) ) {
                if ( is_object( $item_attr_sets->ItemAttributeSet ) ) {
                    $item_attr_sets->ItemAttributeSet = array( $item_attr_sets->ItemAttributeSet );
                }

                foreach ( $item_attr_sets->ItemAttributeSet as $item_attr_set ) {
                    WCLSI_Item_Attributes::insert_or_update_item_attribute_set( $item_attr_set );
                }
            }
        }

        /**
         * Manually sync a product via the "Manual Sync" button in the LightSpeed metabox
         * @see self::render_meta_box()
         */
        function manual_prod_update_ajax() {
            wclsi_verify_nonce();

            if ( ! isset( $_POST['prod_id'] ) ) {
                header( "HTTP/1.0 409 " . __( 'Could not find product ID.', 'woocommerce-lightspeed-pos' ) );
                exit;
            }

            $prod_id = (int) $_POST['prod_id'];
            $ls_prod = new WCLSI_Lightspeed_Prod();

            $ls_prod->init_via_wc_prod_id( $prod_id );

            if ( $ls_prod->id > 0 ) {
                //button cache is to check if update is from import or edit product page for archived prod removal
                wp_cache_add( 'manual_prod_update_button', true, WCLSI_CACHE );
                wp_cache_add( 'manual_prod_update', true, WCLSI_CACHE );
                $msg = WC_LS_Import_Table::process_single_update( $ls_prod->id );


                wp_cache_delete( 'manual_prod_update_button', WCLSI_CACHE );
                if ( $msg['type'] == 'error' ) {
                    header( "HTTP/1.0 409 " . $msg['msg'] );
                    exit;
                } else {
                    wp_send_json(
                        array(
                            'success'   => true,
                            'errors'    => get_settings_errors( 'wclsi_settings' ),
                            'WAIT_TIME' => wclsi_get_api_wait_time_ms()
                        )
                    );
                }
            } else {
                header( "HTTP/1.0 409 " . __( 'Could not find a product ID to update!', 'woocommerce-lightspeed-pos' ) );
                exit;
            }
            exit;
        }

        function set_prod_sync_ajax() {
            wclsi_verify_nonce();

            if ( ! isset( $_POST['prod_id'] ) ) {
                header( "HTTP/1.0 409 " . __( 'Could not find product ID.', 'woocommerce-lightspeed-pos' ) );
                exit;
            }

            if ( ! isset( $_POST['sync_flag'] ) ) {
                header( "HTTP/1.0 409 " . __( 'Could not find sync flag parameter.', 'woocommerce-lightspeed-pos' ) );
                exit;
            }

            $sync_flag = $_POST['sync_flag'] == 'true' ? true : false;
            $prod_id   = (int) $_POST['prod_id'];

            $ls_prod = new WCLSI_Lightspeed_Prod();
            $ls_prod->init_via_wc_prod_id( $prod_id );

            if ( $ls_prod->id > 0 && $ls_prod->wc_prod_id > 0 ) {
                $variations = $ls_prod->variations;
                if ( ! empty( $variations ) ) {
                    foreach ( $variations as $variation ) {
                        update_post_meta( $variation->wc_prod_id, WCLSI_SYNC_POST_META, $sync_flag );
                    }
                }

                $success = update_post_meta( $prod_id, WCLSI_SYNC_POST_META, $sync_flag );
                wp_send_json( array( 'success' => $success ) );
            }

            exit;
        }

        /**
         * Sets 'wclsi_is_synced' in the wclsi_items table when $meta_key = WCLSI_SYNC_POST_META
         *
         * @param $meta_id
         * @param $object_id
         * @param $meta_key
         * @param $_meta_value
         */
        function set_sync_value( $meta_id, $object_id, $meta_key, $_meta_value ) {
            if ( $meta_key == WCLSI_SYNC_POST_META ) {
                global $wpdb, $WCLSI_ITEM_TABLE;

                $wpdb->update(
                    $WCLSI_ITEM_TABLE,
                    array( 'wclsi_is_synced' => $_meta_value ),
                    array( 'wc_prod_id' => $object_id ),
                    array( '%d' ),
                    array( '%d' )
                );
            }
        }

        /*******************
         * Private Methods *
         *******************/

        /**
         * @param WCLSI_Lightspeed_Prod $wclsi_item
         * @param int $parent_id
         *
         * @return WC_Product_Simple|WC_Product_Variable|WC_Product_Variation
         * @throws Exception
         */
        private function init_wc_product( WCLSI_Lightspeed_Prod $wclsi_item, $parent_id = 0 ) {
            global $WCLSI;

            $wc_product = null;
            if ( $wclsi_item->is_matrix_product() ) {
                $wc_product = new WC_Product_Variable();
            } elseif ( $wclsi_item->is_variation_product() ) {
                $wc_product = new WC_Product_Variation();
            } elseif ( $wclsi_item->is_simple_product() ) {
                $wc_product = new WC_Product_Simple();
            }

            if ( is_null( $wc_product ) ) {
                $msg = 'Could not initialize Woo product.';
                WCLSI_Logger::log_error( $msg, $wclsi_item, true );
                throw new Exception( $msg );
            }

            if ( $wclsi_item->is_variation_product() ) {
                $wc_product->set_parent_id( $parent_id );
            }

            // Variation products should inherit status from parent?
            if ( ! $wclsi_item->is_variation_product() ) {
                if ( isset( $WCLSI->settings['wclsi_import_status'] ) &&
                     $WCLSI->settings['wclsi_import_status'] == 'publish' ) {
                    $wc_product->set_status( 'publish' );
                } else {
                    $wc_product->set_status( 'draft' );
                }
            }

            return $wc_product;
        }

        /**
         * @param WC_Product $wc_product
         * @param WCLSI_Lightspeed_Prod $wclsi_item
         *
         * @throws WC_Data_Exception
         */
        private function set_wc_prod_values( WC_Product &$wc_product, WCLSI_Lightspeed_Prod $wclsi_item ) {
            global $WCLSI;
            $selective_sync = $WCLSI->settings[ WCLSI_WC_SELECTIVE_SYNC ];
            $is_update      = $wc_product->get_id() > 0;
            $errors         = [];
            $set_wc_prop    =
                function ( $prop, $arg ) use ( &$wc_product, &$selective_sync, &$is_update, &$errors ) {
                    if ( is_null( $arg ) ) {
                        return false;
                    }

                    try {
                        if ( ! $is_update ) {
                            return $wc_product->{"set_$prop"}( $arg );
                        } elseif ( $is_update && ! empty( $selective_sync[ $prop ] ) ) {
                            return $wc_product->{"set_$prop"}( $arg );
                        }
                    } catch ( Exception $e ) {
                        $errors[] = "Could not update {$prop} for \"{$wc_product->get_formatted_name()}\": <b>{$e->getMessage()}</b>";
                    }

                    return true;
                };

            $set_wc_prop( 'name', $wclsi_item->description );

            $sku = wclsi_get_ls_sku( $wclsi_item );
            if ( isset( $selective_sync['sku'] ) && $this->is_valid_sku( $wclsi_item, $sku, $is_update ) ) {
                $set_wc_prop( 'sku', $sku );
            }

            $set_wc_prop( 'regular_price', $wclsi_item->get_regular_price() );
            $set_wc_prop( 'price', $wclsi_item->get_regular_price() );

            $set_empty_on_zero_sale_price =
                array_key_exists( 'wclsi_empty_on_zero_sale_price', $WCLSI->settings ) &&
                'true' == $WCLSI->settings['wclsi_empty_on_zero_sale_price'];

            $sale_price = $wclsi_item->get_sale_price();
            if ( '0.00' == $sale_price && $set_empty_on_zero_sale_price ) {
                $set_wc_prop( 'sale_price', '' );
            } else {
                $set_wc_prop( 'sale_price', $sale_price );
            }

            $item_e_commerce = $wclsi_item->get_item_e_commerce();
            if ( ! empty( $item_e_commerce ) ) {
                $set_wc_prop( 'description', $item_e_commerce->long_description );
                $set_wc_prop( 'short_description', $item_e_commerce->short_description );
                $set_wc_prop( 'weight', $this->round_item_e_commerce( $item_e_commerce->weight ) );
                $set_wc_prop( 'length', $this->round_item_e_commerce( $item_e_commerce->length ) );
                $set_wc_prop( 'width', $this->round_item_e_commerce( $item_e_commerce->width ) );
                $set_wc_prop( 'height', $this->round_item_e_commerce( $item_e_commerce->height ) );
            }

            $stock_quantity = wclsi_get_lightspeed_inventory( $wclsi_item, true );

            if ( ! $wclsi_item->is_matrix_product() ) {
                $set_wc_prop( 'stock_quantity', $stock_quantity );
            }

            if ( ! $wc_product->is_type( 'variable' ) ) {
                // Don't update stock status unless selective sync stock_quantity is enabled
                if ( ! empty( $selective_sync['stock_quantity'] ) && $wc_product->managing_stock() ) {
                    set_wc_product_stock_status( $wc_product, $stock_quantity );
                }
            }

            // Cache call is set after linking a product
            $force_manage_stock = wp_cache_get( 'wclsi_force_manage_stock', WCLSI_CACHE );
            if ( ! $is_update || $force_manage_stock ) {
                $wc_product->set_catalog_visibility( 'visible' );
                $manage_stock = $wclsi_item->is_matrix_product() ? 'no' : 'yes';
                $wc_product->set_manage_stock( $manage_stock );
            }

            if ( ! empty( $errors ) ) {
                if ( is_admin() ) {
                    add_settings_error(
                        'wclsi_settings',
                        'wclsi_set_wc_prod_value_error',
                        join( ', ', $errors )
                    );
                }

                WCLSI_Logger::log_error(
                    strip_tags( join( ', ', $errors ) ),
                    array(
                        'wc_prod'        => strip_tags( $wc_product->get_formatted_name() ),
                        'wc_prod_id'     => $wc_product->get_id(),
                        'item_id'        => $wclsi_item->item_id,
                        'item_matrix_id' => $wclsi_item->item_matrix_id
                    ),
                    true
                );
            }
        }

        private function round_item_e_commerce( $item_ecommerce_value ) {
            return round( $item_ecommerce_value, 2 );
        }

        private function is_valid_sku( $ls_prod, $sku, $is_update = false ) {
            if ( wclsi_is_matrix_product( $ls_prod ) || $is_update ) {
                return true;
            }

            $post_id = wc_get_product_id_by_sku( $sku );
            if ( $post_id > 0 ) {
                if ( function_exists( 'add_settings_error' ) ) {
                    add_settings_error(
                        'wclsi_settings',
                        'existing_sku',
                        __(
                            sprintf
                            (
                                'Could not set SKU on product "%s": SKU value "%s" already exists.',
                                $ls_prod->description,
                                $sku
                            ),
                            'woocommerce-lightspeed-pos'
                        ),
                        'error'
                    );
                }

                return false;
            } else {
                return true;
            }
        }

        // @todo move this into WCLSI_Lightspeed_Product
        private function persist_import_data( WCLSI_Lightspeed_Prod $item, WC_Product $wc_prod, $sync = false ) {
            global $wpdb, $WCLSI_ITEM_TABLE;

            $data = array(
                'wclsi_import_date' => current_time( 'mysql' ),
                'wc_prod_id'        => $wc_prod->get_id()
            );

            $data_format = array( '%s', '%d' );

            if ( $sync ) {
                $data['wclsi_is_synced'] = true;
                $data_format[]           = '%d';
            }

            $wpdb->update(
                $WCLSI_ITEM_TABLE,
                $data,
                array( 'id' => $item->id ),
                $data_format,
                array( '%d' )
            );

            if ( $item->is_matrix_product() ) {
                $wc_prod->update_meta_data( WCLSI_MATRIX_ID_POST_META, $item->item_matrix_id );
            } elseif ( $item->is_simple_product() ) {
                $wc_prod->update_meta_data( WCLSI_SINGLE_ITEM_ID_POST_META, $item->item_id );
            } elseif ( $item->is_variation_product() ) {
                $wc_prod->update_meta_data( WCLSI_SINGLE_ITEM_ID_POST_META, $item->item_id );
            }

            wp_cache_add( 'wclsi_updating_item', true, WCLSI_CACHE );
            $wc_prod->save();
            wp_cache_delete( 'wclsi_updating_item', WCLSI_CACHE );
        }


        private function create_variations( $variations, $sync, $parent_id ) {
            global $WCLSI;

            //if there are enough variations that it is likely to cause a timeout error, schedule the variation creation
            if ( count( $variations ) > WCLSI_VARIATION_LIMIT ) {
                foreach ( $variations as $ls_variation ) {
                    $args = array(
                        $ls_variation->id,
                        $sync,
                        $parent_id
                    );
                    as_schedule_single_action(
                        time(),
                        'wclsi_create_variation',
                        $args,
                        'wclsi' );
                }
            } else {
                foreach ( $variations as $ls_variation ) {
                    $this->create_variation( $ls_variation->id, $sync, $parent_id );
                }
            }
        }

        public function create_variation( $ls_variation_id, $sync, $parent_id ) {
            global $WCLSI;
            $ls_variation = new WCLSI_Lightspeed_Prod( $ls_variation_id );
            $ls_variation = apply_filters( 'wclsi_create_ls_data_variation', $ls_variation );
            try {
                $wc_variation_prod = $this->init_wc_product( $ls_variation, $parent_id );
            } catch ( Exception $err ) {
                WCLSI_Logger::log_error( "Failed to create variation for product $parent_id", $ls_variation );
            }

            $this->set_wc_prod_values( $wc_variation_prod, $ls_variation );
            $post_id = $wc_variation_prod->save();

            if ( ! is_wp_error( $post_id ) ) {

                // Get & set the category
                $this->handle_product_taxonomy( $ls_variation, $post_id );
                WCLSI_Item_Attributes::set_attributes_for_wc_variation( $wc_variation_prod, $ls_variation );

                // Add a sync flag so our cron-job will know to sync this product
                if ( $sync ) {
                    update_post_meta( $post_id, WCLSI_SYNC_POST_META, true );
                }

                if ( ! empty( $WCLSI->settings['wclsi_wc_selective_sync']['images'] ) ) {
                    WCLSI_Item_Image::enqueue_item_images_to_download( $wc_variation_prod->get_id(), $ls_variation->images );
                }

                $this->persist_import_data( $ls_variation, $wc_variation_prod, $sync );
            }
        }

        private function handle_product_taxonomy( $prod, $post_id ) {
            global $WCLSI;

            if ( $prod->category_id > 0 ) {
                global $wpdb, $WCLSI_ITEM_CATEGORIES_TABLE;

                $ls_cat = $wpdb->get_row( "SELECT * FROM $WCLSI_ITEM_CATEGORIES_TABLE WHERE category_id=$prod->category_id" );

                if ( ! is_null( $ls_cat ) && $ls_cat->wc_cat_id > 0 ) {
                    $cat_id    = intval( $ls_cat->wc_cat_id );
                    $cat_ids   = get_ancestors( $cat_id, 'product_cat' );
                    $cat_ids[] = $cat_id;
                    wp_set_object_terms( $post_id, $cat_ids, 'product_cat', true );

                    if ( array_key_exists( WCLSI_REMOVE_DEFAULT_CAT, $WCLSI->settings ) &&
                         'true' === $WCLSI->settings[ WCLSI_REMOVE_DEFAULT_CAT ] ) {

                        $default_cat_id = intval( get_option( 'default_product_cat', 0 ) );

                        // Note: if the product does not have any categories associated to it (meaning $ls_cat is null)
                        // It will still default the product to the default category

                        wp_remove_object_terms( $post_id, $default_cat_id, 'product_cat' );

                    }
                }
            }

            // Set the tags
            if ( ! empty( $prod->tags ) ) {
                $selective_sync = $WCLSI->settings['wclsi_wc_selective_sync'];
                if ( ! isset( $selective_sync['tags'] ) ) {
                    return;
                }
                wp_set_object_terms( $post_id, $prod->tags, 'product_tag', true );
            }
        }


        //returns error if product is deleted.
        private function handle_archived_products( $wclsi_item ) {
            global $WCLSI;

            if ( key_exists( 'wclsi_handle_archived_products', $WCLSI->settings ) ) {
                $archive_setting = $WCLSI->settings['wclsi_handle_archived_products'];
                $wc_prod_id      = $wclsi_item->wc_prod_id;
                $product_name    = $wclsi_item->description;

                switch ( $archive_setting ) {
                    case ( $archive_setting == 'do nothing' ):
                        return;
                    case ( $archive_setting == 'delete' ):

                        //won't delete product if clicking update from edit product page
                        if ( wp_cache_get( "manual_prod_update_button", WCLSI_CACHE ) ) {
                            wp_cache_delete( "manual_prod_update_button", WCLSI_CACHE );
                            wp_cache_add( 'delete_archived_prod_from_edit_product_page', true, WCLSI_CACHE );
                            WCLSI_Logger::log_sync_event( "delete archive prod cache should be added", true );

                            //show admin notice
                            return;
                        }

                        $this->cleanUpProduct( $wc_prod_id );
                        $wclsi_item->delete();
                        $msg = "Archived Lightspeed product removed: ";
                        WCLSI_Logger::log_sync_event( $msg, $product_name );

                        return $wc_prod_id;
                    case ( $archive_setting == 'draft' ):

                        $msg = "Archived Lightspeed product moved to 'Draft' state: ";
                        WCLSI_Logger::log_sync_event( $msg, $product_name );
                        $status = 'draft';
                        $post   = array( 'ID' => $wc_prod_id, 'post_status' => $status );
                        wp_update_post( $post );

                        return;

//					case ($archive_setting == 'trash bin'):
//						//move product to trash bin
//						return;
                }
            } else {
                //setting key doesn't exist, don't do anything to the product.
                return;
            }
        }


        //@todo How to update attributes if they've been changed in Lightspeed?
        private function update_matrix_item( WCLSI_Lightspeed_Prod $ls_matrix_item, $update_variations = true ) {
            global $WCLSI;
            $post_id = isset( $ls_matrix_item->wc_prod_id ) ? $ls_matrix_item->wc_prod_id : 0;

            if ( $ls_matrix_item->archived ) {
                $result = $this->handle_archived_products( $ls_matrix_item );
                if ( $result > 0 ) {
                    return $result;
                }
            }

            $wc_product = wc_get_product( $post_id );
            if ( ! empty( $wc_product ) ) {
                $wc_product = apply_filters( 'wclsi_update_matrix_item', $wc_product, $ls_matrix_item );

                $this->set_wc_prod_values( $wc_product, $ls_matrix_item );

                // Get & set the category
                $this->handle_product_taxonomy( $ls_matrix_item, $post_id );

                if ( ! empty( $WCLSI->settings['wclsi_wc_selective_sync']['images'] ) ) {
                    WCLSI_Item_Image::enqueue_item_images_to_download( $wc_product->get_id(), $ls_matrix_item->images );
                }

                if ( $update_variations ) {
                    $variations = apply_filters(
                        'wclsi_update_variations_matrix_item',
                        $ls_matrix_item->variations,
                        $post_id
                    );

                    if ( ! empty( $variations ) ) {
                        $this->update_matrix_variations( $variations, $wc_product );
                    }
                }

                delete_transient( 'wc_product_children_' . $post_id );

                wp_cache_add( 'wclsi_updating_item', true, WCLSI_CACHE );
                $wc_product_id = $wc_product->save();
                wp_cache_delete( 'wclsi_updating_item', WCLSI_CACHE );
                WCLSI_Logger::log_debug_event( "Triggered save for WC matrix product with ID $wc_product_id" );
            }

            return $post_id;
        }

        // @todo Handle category updates - what happens if $new_ls_prod has a new category?
        public function update_single_item( $ls_prod, $update_attributes = false ) {

            global $WCLSI;

            // if an item is being updated via the action scheduler, we need to instantiate the product
            if ( ! ( $ls_prod instanceof WCLSI_Lightspeed_Prod ) ) {
                $ls_prod = new WCLSI_Lightspeed_Prod( $ls_prod );
            }
            $wc_prod_id = isset( $ls_prod->wc_prod_id ) ? $ls_prod->wc_prod_id : 0;


            if ( $ls_prod->archived ) {
                $result = $this->handle_archived_products( $ls_prod );
                if ( $result > 0 ) {
                    return $result;
                }
            }

            if ( $wc_prod_id > 0 ) {
                $ls_prod = apply_filters( 'wclsi_update_product', $ls_prod, $wc_prod_id );

                $wc_product = wc_get_product( $wc_prod_id );

                if ( $update_attributes && $ls_prod->is_variation_product() ) {
                    WCLSI_Item_Attributes::set_attributes_for_wc_variation( $wc_product, $ls_prod );
                }


                $old_vals = $this->pull_selective_sync_values( $wc_product );

                $this->set_wc_prod_values( $wc_product, $ls_prod );

                wp_cache_add( 'wclsi_updating_item', true, WCLSI_CACHE );
                $wc_product_id = $wc_product->save();
                wp_cache_delete( 'wclsi_updating_item', WCLSI_CACHE );


                WCLSI_Logger::log_debug_event( "Triggered save for WC product with ID $wc_product_id" );

                $new_vals = $this->pull_selective_sync_values( $wc_product );

                // Set the category & tags
                $this->handle_product_taxonomy( $ls_prod, $wc_prod_id );

                // Process images
                if ( ! empty( $WCLSI->settings['wclsi_wc_selective_sync']['images'] ) ) {
                    WCLSI_Item_Image::enqueue_item_images_to_download( $wc_prod_id, $ls_prod->images );
                }

                $is_manual_update = wp_cache_get( 'manual_prod_update', WCLSI_CACHE );


                if ( $is_manual_update ) {
                    $changes = array_diff( $new_vals, $old_vals );

                    if ( ! empty( $changes ) ) {
                        $msg = "Changes: \nName: {$wc_product->get_name()} \nProduct ID: {$wc_product->get_id()}";
                        WCLSI_Logger::log_sync_event( $msg, $changes );
                    }
                }
            }

            return $wc_prod_id;
        }

        private function pull_selective_sync_values( $wc_prod ): array {
            global $WCLSI;

            $sync_settings = $WCLSI->settings['wclsi_wc_selective_sync'];
            $vals          = array();

            $vals['name']           = $sync_settings['name'] ? $wc_prod->get_name() : null;
            $vals['sku']            = $sync_settings['sku'] ? $wc_prod->get_sku() : null;
            $vals['regular_price']  = $sync_settings['regular_price'] ? $wc_prod->get_regular_price() : null;
            $vals['stock_quantity'] = $sync_settings['stock_quantity'] ? $wc_prod->get_stock_quantity() : null;

            return $vals;
        }
    }

    global $WCLSI_PRODS;
    $WCLSI_PRODS = new LSI_Import_Products();
endif;
