<?php
if ( ! class_exists( 'WCLSI_Link_WC_Prods' ) ) {

    class WCLSI_Link_WC_Prods {
        const MAX_LINK_RETRY = 20;
        const NOTIFICATION_KEY = 'wclsi_linking_notifications';

        public function __construct() {
            add_action( 'wp_ajax_relink_wc_prod_ajax', array( $this, 'relink_wc_prod_ajax' ) );
            add_action( 'wclsi_link_wc_prod_to_lightspeed', array( $this, 'as_wclsi_link_wc_prod' ), 10, 1 );
            add_action( 'admin_notices', array( $this, 'display_linking_notifications' ) );
        }

        /**
         * Displays notifications relating to sync operations
         */
        function display_linking_notifications() {
            if ( is_admin() ) {
                $wclsi_notifications = get_option( WCLSI_NOTIFICATIONS, array() );

                $notifications = null;
                if ( array_key_exists( self::NOTIFICATION_KEY, $wclsi_notifications ) ) {
                    $notifications = $wclsi_notifications[ self::NOTIFICATION_KEY ];
                }

                if ( ! empty( $notifications ) ) {
                    foreach ( $notifications as $notification ) {
                        $class = "notice notice-{$notification[ 'type' ]} is-dismissible";
                        printf( '<div class="%1$s"><p>%2$s</p></div>', $class, $notification['msg'] );
                        unset( $wclsi_notifications['wclsi_linking_notifications'] );
                    }
                    update_option( WCLSI_NOTIFICATIONS, $wclsi_notifications );
                }

                $this->display_as_pending_linking();
            }
        }

        function display_as_pending_linking() {
            /**
             * Look up pending sync events
             */
            $pending_linking_events = as_get_scheduled_actions(
                array(
                    'group'    => 'wclsi',
                    'hook'     => 'wclsi_link_wc_prod_to_lightspeed',
                    'status'   => ActionScheduler_Store::STATUS_PENDING,
                    'per_page' => - 1
                )
            );

            if ( ! empty( $pending_linking_events ) ) {
                $class         = "notice notice-success is-dismissible";
                $pending_prods = '';
                foreach ( $pending_linking_events as $event ) {
                    $args    = $event->get_args();
                    $wc_prod = wc_get_product( $args['wc_prod_id'] );
                    if ( ! empty( $wc_prod ) ) {
                        $permalink     = admin_url( "tools.php?page=action-scheduler&s={$wc_prod->get_id()}&orderby=schedule&order=desc&status=pending" );
                        $pending_prods .= "<li><a href='{$permalink}'>{$wc_prod->get_formatted_name()}</a></li>";
                    }
                }

                printf(
                    '<div class="%1$s">%5$s<p>%2$s</p>%3$s%4$s</div>',
                    $class,
                    'Attempting to link Woo Products to Lightspeed ... these events are being processed in the background.',
                    '<p>Click on a product to see if the action scheduler completed the linking process.</p>',
                    "<ul style='list-style: circle; margin-left: 55px'>$pending_prods</ul>",
                    "<span class='spinner wclsi-spinner' style='float: left; margin-top: 5px'>&nbsp;</span>"
                );
            }
        }

        /**
         * Will attempt to re-link a woocommerce
         */
        function relink_wc_prod_ajax() {
            wclsi_verify_nonce();

            $wc_prod_id = isset( $_POST['wc_prod_id'] ) ? (int) $_POST['wc_prod_id'] : false;

            if ( $wc_prod_id > 0 ) {
                $result = $this->relink_wc_prod( $wc_prod_id );
                if ( is_wp_error( $result ) ) {
                    echo json_encode(
                        array(
                            'success' => false,
                            'errors'  => array(
                                array( 'message' => join( ',', $result->get_error_messages() ) )
                            )
                        )
                    );
                    exit;
                } else {
                    $wc_prod = wc_get_product( $wc_prod_id );
                    $this->handle_linking_notification( $result, $wc_prod_id );
                    $this->update_wc_prod_after_linking( $wc_prod );
                    echo json_encode( array( 'success' => (bool) $result, 'prod_id' => $wc_prod_id ) );
                    exit;
                }
            } else {
                header( "HTTP/1.0 409 " . sprintf( __( 'Product with ID %d does not exist!', 'woocommerce-lightspeed-pos' ), $prod_id ) );
                exit;
            }
        }

        /**
         * @param     $wc_prod_id
         * @param int $retry_count
         *
         * @return bool
         * @throws Exception
         */
        function as_wclsi_link_wc_prod( $wc_prod_id, $retry_count = 0 ): bool {
            // If the bucket level is too high, return early and re-schedule the update to a future time
            if ( wclsi_get_bucket_status() > 0.8 ) {
                $retry_reason = new WP_Error( 'bucket_level_exceeded_0.8', 'Lightspeed bucket level exceeded 0.8' );
                $this->retry_linking( $wc_prod_id, $retry_count, $retry_reason );

                return false;
            }

            $result = $this->relink_wc_prod( $wc_prod_id );
            $this->handle_linking_notification( $result, $wc_prod_id );

            if ( is_wp_error( $result ) ) {
                $err_messages = join( ",", $result->get_error_messages() );
                throw new Exception( $err_messages );
            } else {
                $wc_prod = wc_get_product( $wc_prod_id );
                if ( ! empty( $wc_prod ) ) {
                    $this->update_wc_prod_after_linking( $wc_prod );
                }
            }

            return true;
        }

        private function update_wc_prod_after_linking( $wc_prod ) {
            $wclsi_item = new WCLSI_Lightspeed_Prod();
            $wclsi_item->init_via_wc_prod_id( $wc_prod->get_id() );
            $update_error = null;

            if ( $wclsi_item->id > 0 ) {

                wp_cache_add( 'wclsi_force_manage_stock', true, WCLSI_CACHE );
                $update_result = WC_LS_Import_Table::process_single_update( $wclsi_item->id );
                wp_cache_delete( 'wclsi_force_manage_stock', WCLSI_CACHE );

                if ( 'error' === $update_result['type'] ) {
                    $update_error = new WP_Error(
                        'wclsi_update_error',
                        $update_result['msg'],
                        array( 'ls_mysql_id' => $wclsi_item, '$wc_prod' => $wc_prod )
                    );
                }
            } else {
                $update_error = new WP_Error(
                    'wclsi_update_error',
                    "Could not update product '{$wc_prod->get_formatted_name()}' after link action: " .
                    "wclsi_item was not instantiated successfully",
                    array( 'wc_prod' => $wc_prod )
                );
            }

            if ( ! is_null( $update_error ) ) {
                $this->handle_linking_notification( $update_error, $wc_prod );
            }
        }

        /**
         * @param $linking_result
         * @param $wc_prod_id
         */
        private function handle_linking_notification( $linking_result, $wc_prod_id ) {
            $wclsi_notifications   = get_option( WCLSI_NOTIFICATIONS, array() );
            $linking_notifications = $wclsi_notifications[ self::NOTIFICATION_KEY ] ?: [];
            $wc_prod               = wc_get_product( $wc_prod_id );

            if ( is_wp_error( $linking_result ) ) {
                $err_messages = join( ",", $linking_result->get_error_messages() );
                if ( $wc_prod ) {
                    $linking_notifications[] = array(
                        'msg'  => $err_messages,
                        'type' => 'error'
                    );
                }
            } else {
                if ( $wc_prod ) {
                    $linking_notifications[] = array(
                        'msg'  => "Product '{$wc_prod->get_formatted_name()}' successfully linked!",
                        'type' => 'success'
                    );
                }
            }

            $wclsi_notifications[ self::NOTIFICATION_KEY ] = $linking_notifications;
            update_option( WCLSI_NOTIFICATIONS, $wclsi_notifications );
        }

        private function retry_linking( $wc_prod_id, $retry_count, WP_Error $retry_reason ) {
            if ( $retry_count > self::MAX_LINK_RETRY ) {
                // Action scheduler should mark this as a failed sync
                throw new Exception( $retry_reason->get_error_message() );
            }

            $retry_count ++;

            as_schedule_single_action(
                time() + (int) ( wclsi_get_api_wait_time_ms() / 1000 ),
                'wclsi_link_wc_prod_to_lightspeed',
                array( 'wc_prod_id' => $wc_prod_id, 'retry_count' => $retry_count ),
                'wclsi'
            );

            if ( wclsi_debug() ) {
                $level = wclsi_get_bucket_status();
                $msg   = "Retrying - retry count: $retry_count - bucket level: $level";
                WCLSI_Logger::log_debug_event( $msg );
            }
        }

        /**
         * @param $wc_prod_id
         *
         * @return bool|WP_Error
         */
        function relink_wc_prod( $wc_prod_id ) {

            $wc_prod = wc_get_product( $wc_prod_id );
            if ( ! empty( $wc_prod ) ) {
                if ( $wc_prod->is_type( 'simple' ) ) {
                    return $this->relink_simple_prod( $wc_prod );
                } else if ( $wc_prod->is_type( 'variable' ) ) {
                    return $this->relink_matrix_prod( $wc_prod );
                }
            }

            return new WP_Error(
                'wclsi_cannot_relink',
                __( 'Could not find a WooCommerce product ID to relink with.', 'woocommerce-lightspeed-pos' ),
                $wc_prod_id
            );
        }

        /**
         * @param WC_Product_Simple|false $wc_prod
         *
         * @return bool|WP_Error
         */
        function relink_simple_prod( WC_Product_Simple $wc_prod ) {
            $ls_item_id                 = get_post_meta( $wc_prod->get_id(), WCLSI_SINGLE_ITEM_ID_POST_META, true );
            $wclsi_id                   = WCLSI_Lightspeed_Prod::get_mysql_id( $ls_item_id, 0 );
            $prod_type_mismatch_err_msg =
                "Could not link product '{$wc_prod->get_formatted_name()}'. " .
                "You cannot link a WooCommerce simple product with a Lightspeed variation product.";

            if ( $wclsi_id > 0 ) {
                $wclsi_item = new WCLSI_Lightspeed_Prod( $wclsi_id );
                if ( $wclsi_item->is_variation_product() ) {
                    return new WP_Error( 'wclsi_link_prod_type_mismatch', $prod_type_mismatch_err_msg );
                }

                return $this->relink_wclsi_item( $wclsi_id, $wc_prod );
            } else {
                global $WCLSI, $WCLSI_SINGLE_LOAD_RELATIONS;
                $search_params = array(
                    'load_relations' => json_encode( $WCLSI_SINGLE_LOAD_RELATIONS )
                );

                $sku = $wc_prod->get_sku();
                if ( empty( $sku ) ) {
                    return new WP_Error(
                        'wclsi_linking_sku_missing',
                        "Could not link '{$wc_prod->get_formatted_name()}' - the SKU value is missing!"
                    );
                }

                if ( empty( $ls_item_id ) ) {
                    $search_params['or'] = "customSku=$sku|systemSku=$sku";
                    $ls_item_id          = '';
                }

                $result =
                    $WCLSI->make_api_call(
                        "Account/$WCLSI->ls_account_id/Item/$ls_item_id",
                        "Read",
                        $search_params
                    );

                if ( is_wp_error( $result ) ) {
                    return $result;
                } elseif ( 0 == $result->{'@attributes'}->count ) {
                    return new WP_Error(
                        'wclsi_could_not_find_ls_product',
                        __(
                            "Could not relink! Lightspeed product with SKU: '$sku' could not be found.",
                            'woocommerce-lightspeed-pos'
                        ),
                        'error' );
                } elseif ( $result->{'@attributes'}->count > 1 && ! empty( $sku ) ) {
                    return new WP_Error(
                        'wclsi_could_not_match_ls_product',
                        __(
                            "Could not relink! There was more than one matching Lightspeed product with sku: '$sku'.",
                            'woocommerce-lightspeed-pos'
                        ),
                        'error' );
                } else {
                    $ls_api_item = $result->Item;
                    if ( $ls_api_item->itemID && $ls_api_item->itemMatrixID ) {
                        return new WP_Error( 'wclsi_link_prod_type_mismatch', $prod_type_mismatch_err_msg );
                    }

                    return $this->relink_ls_api_item( $ls_api_item, $wc_prod, true );
                }
            }
        }

        /**
         * @param WC_Product_Variable|false $wc_prod
         *
         * @return WP_Error|bool
         */
        function relink_matrix_prod( WC_Product_Variable $wc_prod ) {
            $ls_matrix_id = $wc_prod->get_meta( WCLSI_MATRIX_ID_POST_META );
            if ( empty( $ls_matrix_id ) ) {
                /**
                 * This is a special case for variable products - Lightspeed matrix products don't have
                 * SKU values. We can work around this by asking the user to input the matrix ID into the SKU field in
                 * WooCommerce.
                 */
                $ls_matrix_id = (int) $wc_prod->get_sku();
            }

            if ( empty( $ls_matrix_id ) ) {
                $ls_matrix_id = $this->get_matrix_id_from_variations( $wc_prod );
            }

            if ( is_wp_error( $ls_matrix_id ) ) {
                return $ls_matrix_id;
            }

            if ( 0 === $ls_matrix_id ) {
                return new WP_Error(
                    'wclsi_linking_item_matrix_id_not_found',
                    "Could not link product '{$wc_prod->get_formatted_name()}' - item_matrix_id not found."
                );
            }

            /**
             * Keep it simple and relink via the API to pull in the most up to date version.
             * We do a check in the DB before calling relink_api_item() anyway.
             */
            $matrix_link_result = $this->relink_matrix_prod_via_ls_api( $wc_prod, $ls_matrix_id );
            if ( is_wp_error( $matrix_link_result ) ) {
                return $matrix_link_result;
            }

            return $this->relink_variation_prods( $wc_prod, $ls_matrix_id );
        }

        /**
         * @param WC_Product_Variable $wc_prod
         * @param                     $ls_matrix_id
         *
         * @return bool|WP_Error
         */
        private function relink_matrix_prod_via_ls_api( WC_Product_Variable $wc_prod, $ls_matrix_id ) {
            global $WCLSI, $WCLSI_MATRIX_LOAD_RELATIONS;
            $search_params = array(
                'load_relations' => json_encode( $WCLSI_MATRIX_LOAD_RELATIONS )
            );

            // Make an API call to retrieve the matrix object
            $result =
                $WCLSI->make_api_call(
                    "Account/$WCLSI->ls_account_id/ItemMatrix/$ls_matrix_id",
                    "Read",
                    $search_params
                );

            if ( is_wp_error( $result ) ) {
                return $result;
            } elseif ( 0 == $result->{'@attributes'}->count ) {
                return new WP_Error(
                    'wclsi_could_not_find_ls_product',
                    __(
                        'Could not relink product! The product was not found in Lightspeed.',
                        'woocommerce-lightspeed-pos'
                    ),
                    'error' );
            } else {
                $ls_api_matrix_item = $result->ItemMatrix;

                return $this->relink_ls_api_item( $ls_api_matrix_item, $wc_prod, true );
            }
        }

        /**
         * @param WC_Product_Variable $wc_prod
         *
         * @return int|WP_Error
         */
        private function get_matrix_id_from_variations( WC_Product_Variable $wc_prod ) {
            $variation_ids = $wc_prod->get_children();
            if ( empty( $variation_ids ) ) {
                return 0;
            }

            $sku_to_lookup = '';
            foreach ( $variation_ids as $variation_id ) {
                $variation = new WC_Product_Variation( $variation_id );
                if ( $variation->get_sku() ) {
                    $sku_to_lookup = $variation->get_sku();
                    break;
                }
            }

            if ( '' === $sku_to_lookup ) {
                return 0;
            }

            global $WCLSI, $WCLSI_SINGLE_LOAD_RELATIONS;
            $search_params = array(
                'load_relations' => json_encode( $WCLSI_SINGLE_LOAD_RELATIONS ),
                'or'             => $this->get_relink_lookup_params( array( $sku_to_lookup ), [] )
            );

            $result = $WCLSI->make_api_call( "Account/$WCLSI->ls_account_id/Item", "Read", $search_params );
            if ( is_wp_error( $result ) ) {
                return $result;
            }

            if ( 0 == $result->{'@attributes'}->count ) {
                return 0;
            }

            return (int) $result->Item->itemMatrixID;
        }

        /**
         * @param WC_Product_Variable $wc_parent_prod
         * @param                     $ls_matrix_id
         *
         * @return bool|WP_Error
         */
        function relink_variation_prods( WC_Product_Variable $wc_parent_prod, $ls_matrix_id ) {
            $variation_ids           = $wc_parent_prod->get_children();
            $wclsi_variation_mapping = [];
            $ls_items_to_lookup      = [];
            $skus_to_lookup          = [];
            $results                 = [];

            $sync = get_post_meta( $wc_parent_prod->get_id(), WCLSI_SYNC_POST_META, true );

            // Try to get the LS item_id via post_meta first
            foreach ( $variation_ids as $wc_variation_id ) {
                $ls_item_id = get_post_meta( $wc_variation_id, WCLSI_SINGLE_ITEM_ID_POST_META, true );

                // If there is no id, try and use the SKU
                if ( empty( $ls_item_id ) ) {
                    $wc_variation                       = wc_get_product( $wc_variation_id );
                    $skus_to_lookup[ $wc_variation_id ] = $wc_variation->get_sku();
                } else {
                    $wclsi_variation_mapping[ $wc_variation_id ] = $ls_item_id;
                }
            }

            if ( ! empty( $wclsi_variation_mapping ) ) {
                foreach ( $wclsi_variation_mapping as $wc_variation_id => $ls_item_id ) {
                    // See if the product exists in wp_wclsi_items table first and try and relink it
                    $wclsi_id = WCLSI_Lightspeed_Prod::get_mysql_id( $ls_item_id, $ls_matrix_id );
                    if ( $wclsi_id > 0 ) {
                        $results[] = $this->relink_wclsi_item( $wclsi_id, wc_get_product( $wc_variation_id ), $sync );
                    } else {
                        $ls_items_to_lookup[ $wc_variation_id ] = $ls_item_id;
                    }
                }
            }

            // Lookup the items we couldn't find
            if ( ! empty( $skus_to_lookup ) || ! empty( $ls_items_to_lookup ) ) {
                global $WCLSI, $WCLSI_SINGLE_LOAD_RELATIONS;

                $search_params = array(
                    'load_relations' => json_encode( $WCLSI_SINGLE_LOAD_RELATIONS ),
                    'itemMatrixID'   => $ls_matrix_id,
                    'or'             => $this->get_relink_lookup_params( $skus_to_lookup, $ls_items_to_lookup )
                );

                $result = $WCLSI->make_api_call( "Account/$WCLSI->ls_account_id/Item", "Read", $search_params );

                if ( is_wp_error( $result ) ) {
                    return $result;
                } elseif ( 0 == $result->{'@attributes'}->count ) {
                    $skus = join( ',', $skus_to_lookup );

                    return new WP_Error(
                        'wclsi_could_not_find_ls_product',
                        __(
                            "Could not relink product! The product(s) with SKU $skus were not found in Lightspeed.",
                            'woocommerce-lightspeed-pos'
                        ),
                        'error' );
                } else {
                    $ls_api_items = $result->Item;
                    if ( ! is_array( $ls_api_items ) ) {
                        $ls_api_items = array( $result->Item );
                    }

                    $items_to_lookup = array_flip( $ls_items_to_lookup ) + array_flip( $skus_to_lookup );
                    foreach ( $ls_api_items as $ls_api_item ) {
                        if ( isset( $items_to_lookup[ $ls_api_item->itemID ] ) ) {
                            $wc_prod_id = $items_to_lookup[ $ls_api_item->itemID ];
                        } elseif ( isset( $items_to_lookup[ $ls_api_item->systemSku ] ) ) {
                            $wc_prod_id = $items_to_lookup[ $ls_api_item->systemSku ];
                        } elseif ( isset( $items_to_lookup[ $ls_api_item->customSku ] ) ) {
                            $wc_prod_id = $items_to_lookup[ $ls_api_item->customSku ];
                        }

                        $results[] = $this->relink_ls_api_item( $ls_api_item, wc_get_product( $wc_prod_id ), $sync );
                    }
                }
            }

            foreach ( $results as $result ) {
                if ( is_wp_error( $result ) ) {
                    return $result;
                }
            }

            return true;
        }

        /**
         * "Re-links" a WC product with one that does NOT exist in the wp_wclsi_items table. It achieves this by first
         * inserting a new item in the table and then filling in the "wc_prod_id" column with the WC prod id.
         *
         * @param            $ls_api_item
         * @param WC_Product $wc_prod
         * @param false $sync
         *
         * @return bool|WP_Error
         */
        private function relink_ls_api_item( $ls_api_item, WC_Product $wc_prod, $sync = false ) {
            // last check on the wp_wclsi_items table in case post_meta did not exist for the product to begin with
            $wclsi_mysql_id = WCLSI_Lightspeed_Prod::item_exists( $ls_api_item );
            if ( $wclsi_mysql_id > 0 ) {
                return $this->relink_wclsi_item( $wclsi_mysql_id, $wc_prod );
            }

            $key   = $wc_prod->is_type( 'variable' ) ? WCLSI_MATRIX_ID_POST_META : WCLSI_SINGLE_ITEM_ID_POST_META;
            $value = $wc_prod->is_type( 'variable' ) ? $ls_api_item->itemMatrixID : $ls_api_item->itemID;
            update_post_meta( $wc_prod->get_id(), $key, $value );

            // For variation items in case the parent matrix product has sync enabled
            if ( $sync ) {
                $ls_api_item->wclsi_is_synced = true;
                update_post_meta( $wc_prod->get_id(), WCLSI_SYNC_POST_META, true );
            }

            $ls_api_item->wc_prod_id = $wc_prod->get_id();
            $import_date             = $wc_prod->get_date_created();
            if ( ! empty( $import_date ) ) {
                $ls_api_item->wclsi_import_date = date_i18n( 'Y-m-d H:i:s', $import_date->getTimestamp() );
            }

            return WCLSI_Lightspeed_Prod::insert_ls_api_item( $ls_api_item ) > 0;
        }

        /**
         * @param       $wclsi_mysql_id - the id of the WCLSI_Lightspeed object from the wclsi_items table
         * @param       $wc_prod
         * @param false $sync
         *
         * @return bool|WP_Error
         */
        private function relink_wclsi_item( $wclsi_mysql_id, WC_Product $wc_prod, $sync = true ) {
            $wclsi_prod = new WCLSI_Lightspeed_Prod( $wclsi_mysql_id );

            if ( $wclsi_prod->wc_prod_id > 0 ) {
                if ( ! $wc_prod->get_meta( WCLSI_SYNC_POST_META ) || ! $wclsi_prod->wclsi_is_synced ) {
                    $wclsi_prod->update_column( 'wclsi_is_synced', true );
                    update_post_meta( $wc_prod->get_id(), WCLSI_SYNC_POST_META, true );

                    return true;
                } else {
                    $error_msg =
                        __(
                            "Product \"{$wc_prod->get_formatted_name()}\" is already linked to Lightspeed!",
                            'woocommerce-lightspeed-pos'
                        );

                    return new WP_Error( 'wclsi_already_linked', $error_msg );
                }
            }

            $key   = $wclsi_prod->is_matrix_product() ? WCLSI_MATRIX_ID_POST_META : WCLSI_SINGLE_ITEM_ID_POST_META;
            $value = $wclsi_prod->is_matrix_product() ? $wclsi_prod->item_matrix_id : $wclsi_prod->item_id;
            update_post_meta( $wc_prod->get_id(), $key, $value );

            // For variation items in case the parent matrix product has sync enabled
            if ( $sync ) {
                $wclsi_prod->update_column( 'wclsi_is_synced', true );
                update_post_meta( $wc_prod->get_id(), WCLSI_SYNC_POST_META, true );
            }

            $import_date = $wc_prod->get_date_created();
            if ( ! empty( $import_date ) ) {
                $wclsi_prod->update_column( 'wclsi_import_date', date_i18n( 'Y-m-d H:i:s', $import_date->getTimestamp() ) );
            }

            $update_result = $wclsi_prod->update_column( 'wc_prod_id', $wc_prod->get_id() );
            if ( false === $update_result ) {
                global $wpdb;
                $error_msg =
                    __(
                        "Could not relink product, there was a problem with updating the wc_prod_id column. {$wpdb->last_error}",
                        'woocommerce-lightspeed-pos'
                    );

                return new WP_Error( 'wclsi_could_not_relink_mysql', $error_msg, $wclsi_mysql_id );
            } else {
                return true;
            }
        }

        /**
         * @param $skus_to_lookup
         * @param $ls_items_to_lookup
         *
         * @return string
         */
        private function get_relink_lookup_params( $skus_to_lookup, $ls_items_to_lookup ) {
            $sku_lookup_params = '';
            foreach ( $skus_to_lookup as $sku ) {
                if ( $sku !== end( $skus_to_lookup ) ) {
                    $sku_lookup_params .= "systemSku=$sku|customSku=$sku|";
                } else {
                    $sku_lookup_params .= "systemSku=$sku|customSku=$sku";
                }
            }

            $ls_items_lookup_params = '';
            foreach ( $ls_items_to_lookup as $item_id ) {
                if ( $item_id !== end( $ls_items_to_lookup ) ) {
                    $ls_items_lookup_params .= "itemID=$item_id|";
                } else {
                    $ls_items_lookup_params .= "itemID=$item_id";
                }
            }

            if ( empty( $ls_items_lookup_params ) ) {
                return $sku_lookup_params;
            } else {
                return $sku_lookup_params . '|' . $ls_items_lookup_params;
            }
        }
    }

    $WCLSI_Link_WC_Prods = new WCLSI_Link_WC_Prods();
}
