<?php
/*************************************
 * @class LSI_Sync_Sales
 * handles syncing sales to Lightspeed
 */


if ( ! class_exists( 'LSI_Sync_Sales' ) ) :
    class LSI_Sync_Sales {
        /**
         * WooCommerce order to be created or refunded in Lightspeed
         * @var WC_Order
         */
        protected $order;

        /**
         * Refund that is being processed, if applicable
         * @var
         */
        protected $refund;

        /**
         * Items in order or refund that are not linked to Lightspeed
         * @var array
         */
        protected $unlinked_items;

        /**
         * Lightspeed tax category ID
         * @var int
         */
        protected $ls_tax_id;


        function __construct() {
            add_action( 'woocommerce_order_status_processing', array( $this, 'create_sale' ) );
            add_action( 'woocommerce_order_status_cancelled', array( $this, 'refund_sale' ) );
            add_action( 'updated_option', array( $this, 'initialize_sales_sync' ), 10, 3 );
            add_action( 'woocommerce_new_order', array( $this, 'add_syncing_new_order_flag' ) );
            add_action( 'woocommerce_before_order_object_save', array( $this, 'add_syncing_cancelled_order_flag' ) );
            add_action( 'woocommerce_order_fully_refunded', array( $this, 'refund_sale' ) );
            add_action( 'woocommerce_order_partially_refunded', array( $this, 'partially_refund_sale' ) );
            add_action( 'wp_ajax_export_tax_rates', array( $this, 'export_rates_ajax' ) );
            add_action( 'wp_ajax_delete_tax_rates', array( $this, 'delete_tax_categories_in_lookup_table_ajax' ) );
            add_action( 'wclsi_process_tax_category', array( $this, 'process_combined_tax_ids' ), 10, 5 );
            add_action( 'wclsi_delete_tax_category', array(
                $this,
                'delete_tax_category_and_remove_from_lookup_table'
            ), 10, 1 );
            add_action( 'admin_notices', array( $this, 'display_sales_notices' ) );
        }


        function display_sales_notices() {
            if ( is_admin() ) {
                $wclsi_notifications = get_option( WCLSI_NOTIFICATIONS, array() );
                $notification        = null;
                if ( array_key_exists( 'wclsi_sales_sync_notifications', $wclsi_notifications ) ) {
                    $notification = $wclsi_notifications['wclsi_sales_sync_notifications'];
                }

                if ( ! empty ( $notification ) ) {
                    $class = "notice notice-{$notification['type']} is-dismissible";
                    printf( '<div class="%1$s"><p>%2$s</p></div>', $class, $notification['msg'] );

                    unset( $wclsi_notifications['wclsi_sales_sync_notifications'] );
                    update_option( WCLSI_NOTIFICATIONS, $wclsi_notifications );
                }
            }
        }

        function set_sales_notice( $msg, $type ) {
            $wclsi_notifications                                   = get_option( WCLSI_NOTIFICATIONS, array() );
            $wclsi_notifications['wclsi_sales_sync_notifications'] = array(
                'msg'  => $msg,
                'type' => $type
            );
            update_option( WCLSI_NOTIFICATIONS, $wclsi_notifications );
        }


        /**
         * Create a new incomplete sale in Lightspeed. Adds Lightspeed sale ID to woocommerce order post meta
         *
         * @param $order_id
         */
        function create_sale( $order_id ) {
            global $WCLSI;

            if ( ! $this->can_sync_sales( $order_id ) ) {
                return;
            }

            $this->order = wc_get_order( $order_id );

            if ( $this->order->get_meta( '_wclsi_synced_with_ls', true ) ) {
                return;
            }


            $this->set_sales_tax_rate_id();

            $items = $this->order->get_items();

            $sale_lines = $this->build_sale_line( $items );

            $data = $this->build_ls_payload( $sale_lines );

            $result = $WCLSI->make_api_call( 'Account.Sale', 'Create', '', $data );

            if ( ! is_wp_error( $result ) ) {
                $this->order->update_meta_data( 'wclsi_ls_sale_id', $result->Sale->saleID );
                $this->order->update_meta_data( '_wclsi_synced_with_ls', true );
                $this->order->save();

                $msg          = "Successful sync for woocommerce order '$order_id' in Lightspeed";
                $sale_summary = array();

                foreach ( $sale_lines['SaleLine'] as $item_id => $item ) {
                    $line    = array(
                        "Product:"  => $item['name'],
                        "Quantity:" => $item['unitQuantity']
                    );
                    $lines[] = $line;
                }

                $sale_summary['lines'] = $lines;

                //if there are unlinked items in the order, list them
                if ( ! empty( $this->unlinked_items ) ) {
                    $sale_summary['unlinked items omitted'] = $this->unlinked_items;
                }
                WCLSI_Logger::log_sync_event( $msg, $sale_summary );
            } else {
                $msg = "Error syncing woocommerce order '$order_id' in Lightspeed";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );
            }
        }


        /**
         * Refund a sale in Lightspeed. Must be a complete sale including tax
         *
         * @param $order_id
         */
        function refund_sale( $order_id ) {
            global $WCLSI;

            if ( ! $this->can_sync_sales( $order_id ) ) {
                return;
            }

            $this->order = wc_get_order( $order_id );


            $ls_sale_id = $this->order->get_meta( 'wclsi_ls_sale_id', true );

            if ( empty( $ls_sale_id ) ) {
                $msg = "Error refunding woocommerce order '$order_id' in Lightspeed. There is no corresponding Lightspeed sale ID.";
                WCLSI_Logger::log_error( $msg );

                return;
            }

            $data = $this->build_ls_payload();

            $result = $WCLSI->make_api_call( 'Account.Sale.' . $ls_sale_id . '.refund', 'Create', '', $data );
            if ( ! is_wp_error( $result ) ) {
                $this->order->delete_meta_data( '_wclsi_synced_with_ls' );
                $this->order->save();

                $msg = "Successfully refunded woocommerce order '$order_id' in Lightspeed";
                WCLSI_Logger::log_sync_event( $msg );
            } else {
                $msg = "Error refunding woocommerce order '$order_id' in Lightspeed";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );
            }
        }


        /**
         * Refund a partial sale in Lightspeed
         *
         * @param $order_id
         * @param $refund_id
         *
         * @return void
         */
        function partially_refund_sale( $order_id ) {
            global $WCLSI;

            if ( ! $this->can_sync_sales( $order_id ) ) {
                return;
            }

            $this->order = wc_get_order( $order_id );
            $refunds     = $this->order->get_refunds();

            // the first refund on the order should be the most recent. we also set the order to be the refund because we no longer need the original order
            $this->refund = wc_get_order( $refunds[0]->id );

            $this->set_sales_tax_rate_id();

            $refund_items = $this->refund->get_items();

            $items = $this->order->get_items();

            $sale_line = $this->build_sale_line( $refund_items );

            $ls_sale_id = $ls_sale_id = $this->order->get_meta( 'wclsi_ls_sale_id', true );


            $parent_sale_line_ids = $this->fetch_parent_sale_line_ids( $ls_sale_id );

            //if fetching the parent sale line ids failed, abort syncing
            if ( ! $parent_sale_line_ids ) {
                return false;
            }
            //add parent sale line ids to corresponding refund lines
            foreach ( $sale_line["SaleLine"] as $key => $line ) {
                $sale_line["SaleLine"][ $key ]["parentSaleLineID"] = $parent_sale_line_ids[ $line["itemID"] ];
            }


            $data = $this->build_ls_payload( $sale_line );

            $result = $WCLSI->make_api_call( "Account.Sale.{$ls_sale_id}.refund", 'Create', '', $data );

            if ( ! is_wp_error( $result ) ) {
                $msg          = "Successful sync for woocommerce refund '{$this->refund->get_id()}' in Lightspeed";
                $sale_summary = array();

                foreach ( $sale_line['SaleLine'] as $item_id => $item ) {
                    $line    = array(
                        "Product:"  => $item['name'],
                        "Quantity:" => $item['unitQuantity']
                    );
                    $lines[] = $line;
                }

                $sale_summary['lines'] = $lines;

                //if there are unlinked items in the order, list them
                if ( ! empty( $unlinked_items ) ) {
                    $sale_summary['unlinked items omitted'] = $unlinked_items;
                }
                WCLSI_Logger::log_sync_event( $msg, $sale_summary );
            } else {
                $msg = "Error syncing woocommerce refund '{$this->refund->get_id()}' in Lightspeed";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );
            }
        }

        /**
         * Build sale lines from array of items. Also separates any items that are not linked with Lightspeed
         *
         * @param $items
         *
         * @return array[]
         */
        function build_sale_line( $items ): array {
            global $wpdb, $WCLSI_ITEM_TABLE;
            $sale_lines = array();

            foreach ( $items as $item_number => $item ) {
                $product        = $item->get_product();
                $woo_product_id = $product->get_id();
                $lsi_item_id    = $wpdb->get_var( $wpdb->prepare( "SELECT item_id FROM $WCLSI_ITEM_TABLE WHERE wc_prod_id = $woo_product_id" ) );
                $item_name      = $item->get_name();
                $quantity       = $item->get_quantity();

                $active_price  = $product->get_price();
                $regular_price = $product->get_regular_price();

                // check if item is on sale
                if ( $active_price != $regular_price ) {
                    $discount       = $regular_price - $active_price;
                    $total_discount = $discount * $quantity;
                }

                $line = array(
                    "name"         => $item_name,
                    "unitQuantity" => $quantity,
                    "itemID"       => $lsi_item_id,
                    "wooProdID"    => $woo_product_id,
                );

                if ( ! isset( $this->ls_tax_id ) ) {
                    $line["tax"] = "false";
                } else {
                    $line["taxCategoryID"] = $this->ls_tax_id;
                }

                if ( isset( $total_discount ) ) {
                    $line["discountAmount"] = $total_discount;
                    unset( $total_discount );
                }

                //check if item is linked with lightspeed
                $is_synced = get_post_meta( $woo_product_id, WCLSI_SYNC_POST_META, true );
                if ( $is_synced ) {
                    $sale_lines[] = $line;
                } else {
                    $unlinked_items[] = $item;
                }
            }

            return array( "SaleLine" => $sale_lines, "unlinked_items" => $unlinked_items );
        }


        /**
         * Fetches list of Lightspeed employees
         * @return array
         */
        function get_ls_employees(): array {
            global $WCLSI;

            $result = $WCLSI->make_api_call( 'Account.Employee', 'Read' );

            if ( ! is_wp_error( $result ) ) {
                $employees = array();
                if ( is_array( $result->Employee ) ) {
                    foreach ( $result->Employee as $ls_employee ) {
                        $first_name = $ls_employee->firstName;
                        $last_name  = $ls_employee->lastName;

                        if ( is_object( $first_name ) || is_object( $last_name ) ) {
                            $msg      = "Unexpected employee name format";
                            $employee = array(
                                'text'  => $msg,
                                'value' => $ls_employee->employeeID
                            );

                            $data = array(
                                "first_name" => $first_name,
                                "last_name"  => $last_name
                            );
                            WCLSI_Logger::log_error( $msg, $data );
                        } else {
                            $employee = array(
                                'text'  => "$first_name $last_name",
                                'value' => $ls_employee->employeeID
                            );
                        }
                        $employees[] = $employee;
                    }
                } else {
                    $first_name = $result->Employee->firstName;
                    $last_name  = $result->Employee->lastName;

                    if ( is_object( $first_name ) || is_object( $last_name ) ) {
                        $msg      = "Unexpected employee name format";
                        $employee = array(
                            'text'  => $msg,
                            'value' => $result->Employee->employeeID
                        );

                        $data = array(
                            "first_name" => $first_name,
                            "last_name"  => $last_name
                        );
                        WCLSI_Logger::log_error( $msg, $data );
                    } else {
                        $employee = array(
                            'text'  => "$first_name $last_name",
                            'value' => $result->Employee->employeeID
                        );
                    }
                    $employees[] = $employee;
                }

                return $employees;
            } else {
                $msg = "Error fetching employee list";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );

                return array( 'text' => 'Failed to get employees', 'value' => '0' );
            }

        }


        /**
         * Fetches list of Lightspeed registers
         * @return array
         */
        function get_ls_registers(): array {
            global $WCLSI;

            $result = $WCLSI->make_api_call( 'Account.Register', 'Read' );

            if ( ! is_wp_error( $result ) ) {
                $registers = array();
                if ( is_array( $result->Register ) ) {
                    foreach ( $result->Register as $ls_register ) {
                        if ( $ls_register->shopID === $WCLSI->settings[ WCLSI_INVENTORY_SHOP_ID ] ) {
                            $register    = array( 'text' => $ls_register->name, 'value' => $ls_register->registerID );
                            $registers[] = $register;
                        }
                    }
                } else {
                    $registers[] = array( 'text' => $result->Register->name, 'value' => $result->Register->registerID );
                }

                return $registers;
            } else {
                $msg = "Error fetching register list";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );

                return array( 'text' => 'Failed to get registers', 'value' => '0' );
            }
        }


        /**
         * Create or update a tax category in lightspeed
         *
         * @param string $name
         * @param string|float $tax_rate
         * @param bool $tax_inclusive
         * @param bool $create
         * @param int|null $tax_category_id
         *
         * @return string[]|void
         */
        function create_or_update_ls_tax_category( string $name = "Woocommerce", $tax_rate = WCLSI_LS_DEFAULT_TAX_RATE, bool $tax_inclusive = false, bool $create = true, int $tax_category_id = null ) {

            if ( $create ) {
                return $this->create_ls_tax_category( $name, $tax_rate, $tax_inclusive );
            } else {
                $result = $this->update_ls_tax_category( $name, $tax_rate, $tax_inclusive, $tax_category_id );

                if ( $result === false ) {
                    //wait to create new tax category to avoid overloading api
                    sleep( 12 );

                    return $this->create_ls_tax_category( $name, $tax_rate, $tax_inclusive );
                } else {
                    return $result;
                }
            }
        }

        function create_ls_tax_category( string $name = "Woocommerce", $tax_rate = WCLSI_LS_DEFAULT_TAX_RATE, bool $tax_inclusive = false ) {
            global $WCLSI;

            $data = array(
                "tax1Name"       => $name,
                "tax1Rate"       => $tax_rate,
                "isTaxInclusive" => $tax_inclusive
            );

            $action       = 'Create';
            $control_name = "Account.TaxCategory";

            $result = $WCLSI->make_api_call( $control_name, $action, '', $data );

            if ( ! is_wp_error( $result ) ) {
                return $result->TaxCategory->taxCategoryID;
            } else {
                $msg = "Error creating tax category $name";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );

                return false;
            }
        }

        function update_ls_tax_category( string $name = "Woocommerce", $tax_rate = WCLSI_LS_DEFAULT_TAX_RATE, bool $tax_inclusive = false, int $tax_category_id = null ) {
            global $WCLSI;

            $data = array(
                "taxCategoryID"  => $tax_category_id,
                "tax1Name"       => $name,
                "tax1Rate"       => $tax_rate,
                "isTaxInclusive" => $tax_inclusive
            );

            $action       = 'Update';
            $control_name = "Account.TaxCategory.$tax_category_id";

            $result = $WCLSI->make_api_call( $control_name, $action, '', $data );

            if ( ! is_wp_error( $result ) ) {
                return $result->TaxCategory->taxCategoryID;
            } else {
                $msg = "Error updating tax category $name, should default to create";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );

                return false;
            }
        }


        /**
         * Create a tax category in lightspeed and add the ID to WCLSI settings
         *
         * @param string $name
         *
         * @return string[]|void
         */
        function create_ls_payment_type( string $name = "Woocommerce" ) {
            global $WCLSI;

            $data = array(
                "name"             => $name,
                "requireCustomer"  => "false",
                "internalReserved" => "false",
                "type"             => "user defined"
            );

            $result = $WCLSI->make_api_call( 'Account.PaymentType', 'Create', '', $data );


            if ( ! is_wp_error( $result ) ) {
                $payment_type_id                          = $result->PaymentType->paymentTypeID;
                $WCLSI->settings[ WCLSI_PAYMENT_TYPE_ID ] = $payment_type_id;
                update_option( $WCLSI->get_settings_option_key(), $WCLSI->settings );
            } else {
                $msg = "Error creating payment type";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );

                return array( 'text' => 'Failed to create payment type', 'value' => '0' );
            }
        }


        /**
         * Initializes payment type by checking if the named payment type exists in Lightspeed and creates a new one if it does not
         *
         * @param string $name
         *
         * @return void
         */
        function initialize_payment_type( string $name = 'Woocommerce' ) {
            global $WCLSI;

            $result = $WCLSI->make_api_call( 'Account.PaymentType', 'Read' );
            foreach ( $result->PaymentType as $key => $payment_type ) {
                if ( $payment_type->name === $name ) {
                    $WCLSI->settings[ WCLSI_PAYMENT_TYPE_ID ] = $payment_type->paymentTypeID;
                    update_option( $WCLSI->get_settings_option_key(), $WCLSI->settings );

                    return;
                }
            }

            LSI_Sync_Sales::create_ls_payment_type();
        }

        /**
         * checks if sales syncing has been switched from off to on and then initializes the necessary fields
         *
         * @param $option_name
         * @param $old_settings
         * @param $new_settings
         *
         * @return void
         */
        function initialize_sales_sync( $option_name, $old_settings, $new_settings ) {
            global $WCLSI;

            if ( $option_name == $WCLSI->get_settings_option_key() ) {
                if ( ( ! array_key_exists( WCLSI_SALES_SYNC, $old_settings ) || false == $old_settings[ WCLSI_SALES_SYNC ] ) ) {
                    if ( ( array_key_exists( WCLSI_SALES_SYNC, $new_settings ) && true == $new_settings[ WCLSI_SALES_SYNC ] ) ) {
                        LSI_Sync_Sales::initialize_payment_type();
                    }
                }
            }
        }

        /**
         * adds a flag to the cache indicating that a new order is being synced
         * @return void
         */
        function add_syncing_new_order_flag() {
            global $WCLSI;
            if ( ! array_key_exists( WCLSI_SALES_SYNC, $WCLSI->settings ) ||
                 'true' !== $WCLSI->settings[ WCLSI_SALES_SYNC ] ) {
                return;
            }

            wp_cache_add( "wclsi_syncing_sale_or_refund", true, WCLSI_CACHE );
        }

        /**
         * adds a flag to the cache indicating that an order has been cancelled
         * @return void
         */
        function add_syncing_cancelled_order_flag() {
            global $WCLSI;
            if ( ! array_key_exists( WCLSI_SALES_SYNC, $WCLSI->settings ) ||
                 'true' !== $WCLSI->settings[ WCLSI_SALES_SYNC ] || $_GET["action"] !== 'mark_cancelled' ) {
                return;
            }

            wp_cache_add( "wclsi_syncing_sale_or_refund", true, WCLSI_CACHE );
        }

        /**
         * checks if a sale or refund is being synced
         * @return false|mixed
         */
        public static function is_syncing_sale() {
            return wp_cache_get( "wclsi_syncing_sale_or_refund", WCLSI_CACHE );
        }

        /**
         * prevents syncing a sale if sales syncing is turned off or the employee id and register id are not set
         *
         * @param $order_id
         *
         * @return bool
         */
        function can_sync_sales( $order_id ): bool {
            global $WCLSI;

            // if sales syncing isn't turned on, return false
            if ( ! array_key_exists( WCLSI_SALES_SYNC, $WCLSI->settings ) || true != $WCLSI->settings[ WCLSI_SALES_SYNC ] ) {
                return false;
            }

            // if an employee and register are not selected, return false and report an error
            if ( ! array_key_exists( WCLSI_EMPLOYEE_ID, $WCLSI->settings ) || empty( $WCLSI->settings[ WCLSI_EMPLOYEE_ID ] ) ) {
                $msg = "Failed syncing Failed syncing order $order_id : Employee not selected. Check your sales sync settings.";
                WCLSI_Logger::log_error( $msg );

                return false;
            }
            // if an employee and register are not selected, return false and report an error
            if ( ! array_key_exists( WCLSI_REGISTER_ID, $WCLSI->settings ) || empty( $WCLSI->settings[ WCLSI_REGISTER_ID ] ) ) {
                $msg = "Failed syncing order $order_id : Register not selected. Check your sales sync settings.";
                WCLSI_Logger::log_error( $msg );

                return false;
            }

            // if no condition cancels sales syncing, return true
            return true;
        }

        /**
         * calculates the payment by removing the tax and shipping to ensure a balanced sale
         * @return float|string
         */
        function calc_payment() {
            $shipping_total = $this->refund ? $this->refund->get_shipping_total() + $this->refund->get_shipping_tax() : $this->order->get_shipping_total() + $this->order->get_shipping_tax();
            $payment        = $this->refund ? $this->refund->get_total() : $this->order->get_total();
            $tax_total      = $this->refund ? $this->refund->get_total_tax() : $this->order->get_total_tax();
            $shipping_tax   = $this->refund ? $this->refund->get_shipping_tax() : $this->order->get_shipping_tax();

            if ( ! isset( $this->ls_tax_id ) ) {
                return $payment - $shipping_total - $tax_total + $shipping_tax;
            } else {
                return $payment - $shipping_total;
            }
        }

        /**
         * builds a payment line. includes the payment total if syncing a sale or a partial refund
         *
         * @param null $sale_lines
         * @param null $item
         *
         * @return array[][]|array[][][]
         */
        function build_payment_line( $sale_lines = null ): array {
            global $WCLSI;

            // if sale lines are empty, we only need to know the paymentTypeID and not the amount of the payment
            if ( ! empty( $sale_lines ) && ! empty( $this->order ) ) {
                $payment = $this->calc_payment();

                //subtract unlinked items from payment and build array for logging
                if ( isset( $sale_lines['unlinked_items'] ) ) {
                    foreach ( $sale_lines['unlinked_items'] as $item_number => $item ) {
                        $line_total             = isset( $this->refund ) ? $this->refund->get_line_total( $item ) : $this->order->get_line_total( $item );
                        $payment                -= $line_total;
                        $this->unlinked_items[] = array(
                            'Product:'        => $item->get_name(),
                            'Woo Product ID:' => $item->get_product()->get_id(),
                            'Quantity:'       => $item->get_quantity(),
                            'Payment:'        => $line_total
                        );
                    }
                }

                return array(
                    'SalePayment' => array(
                        array(
                            'amount'      => $payment,
                            'PaymentType' => array( 'paymentTypeID' => $WCLSI->settings[ WCLSI_PAYMENT_TYPE_ID ] )
                        )
                    )
                );
            } else {
                return array(
                    'SalePayment' => array(
                        array(
                            'PaymentType' => array( 'paymentTypeID' => $WCLSI->settings[ WCLSI_PAYMENT_TYPE_ID ] )
                        )
                    )
                );
            }


        }

        /**
         * builds the payload to sync to Lightspeed
         *
         * @param $sale_lines
         *
         * @return array
         */
        function build_ls_payload( $sale_lines = null ): array {
            global $WCLSI;

            return array(
                "completed"             => true,
                "employeeID"            => $WCLSI->settings[ WCLSI_EMPLOYEE_ID ],
                "shopID"                => $WCLSI->settings[ WCLSI_INVENTORY_SHOP_ID ],
                "registerID"            => $WCLSI->settings[ WCLSI_REGISTER_ID ],
                "referenceNumber"       => isset( $this->order ) ? (string) $this->order->get_id() : null,
                "referenceNumberSource" => "WooCommerce",
                "SaleLines"             => $sale_lines,
                "SalePayments"          => $this->build_payment_line( $sale_lines ),
            );
        }

        /**
         * fetches parent sale line ids for partial refunds
         *
         * @param $ls_sale_id
         *
         * @return array|false
         */
        function fetch_parent_sale_line_ids( $ls_sale_id ) {
            $result = $this->fetch_sale_with_sale_lines( $ls_sale_id );

            if ( is_wp_error( $result ) ) {
                $msg = "Error fetching sale $ls_sale_id in Lightspeed for refund";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );

                return false;
            }

            $parent_sale_lines = $result->Sale->SaleLines;

            $parent_sale_line_ids = [];

            if ( is_array( $parent_sale_lines->SaleLine ) ) {
                foreach ( $parent_sale_lines->SaleLine as $key => $parent_sale_line ) {
                    $parent_sale_line_ids[ $parent_sale_line->itemID ] = $parent_sale_line->saleLineID;
                }
            } else {
                $parent_sale_line_ids[ $parent_sale_lines->SaleLine->itemID ] = $parent_sale_lines->SaleLine->saleLineID;
            }

            return $parent_sale_line_ids;
        }

        /**
         * fetch a Lightspeed sale with sale lines included
         *
         * @param $ls_sale_id
         *
         * @return array|false|mixed|object|WP_Error
         */
        function fetch_sale_with_sale_lines( $ls_sale_id ) {
            global $WCLSI;

            $data           = array(
                "saleID" => $ls_sale_id
            );
            $load_relations = array(
                "load_relations" => json_encode( array( "SaleLines" ) )
            );

            return $WCLSI->make_api_call( "Account.Sale.$ls_sale_id", 'Read', $load_relations, $data );
        }

        /**
         * Get all tax rates that are set up in woocommerce
         * @return array
         */
        function get_wc_tax_rates(): array {
            $all_tax_rates = [];
            $tax_classes   = WC_Tax::get_tax_classes(); // Retrieve all tax classes.
            if ( ! in_array( '', $tax_classes ) ) { // Make sure "Standard rate" (empty class name) is present.
                array_unshift( $tax_classes, '' );
            }
            foreach ( $tax_classes as $tax_class ) { // For each tax class, get all rates.
                $taxes         = WC_Tax::get_rates_for_tax_class( $tax_class );
                $all_tax_rates = array_merge( $all_tax_rates, $taxes );
            }

            return $all_tax_rates;
        }


        /**
         * Export tax rates from woocommerce to lightspeed and update tax lookup table to establish link between the two
         * @return void
         */


        function export_rates_ajax() {
            wclsi_verify_nonce();
            global $wpdb, $WCLSI_TAX_LOOKUP_TABLE;

            $tax_rates = $this->get_wc_tax_rates();

            $rates_in_buckets = array();

            foreach ( $tax_rates as $tax_rate ) {
                $wc_tax_rate_id                                                      = $tax_rate->tax_rate_id;
                $rate_code                                                           = WC_Tax::get_rate_code( $wc_tax_rate_id );
                $exploded_code                                                       = explode( '-', $rate_code );
                $rates_in_buckets[ $exploded_code[ count( $exploded_code ) - 1 ] ][] = array( $tax_rate );
            }
            foreach ( $rates_in_buckets as $key => $bucket ) {
                $rates_in_buckets[ $key ][] = array();
            }

            $combined_rates         = $this->get_filtered_combined_ids( $rates_in_buckets );
            $tax_categories_created = 0;


            foreach ( $combined_rates as $rates ) {
                $combined_tax_rate  = 0;
                $combined_tax_code  = '';
                $combined_wc_tax_id = '';
                if ( count( $rates ) > 0 ) {
                    foreach ( $rates as $rate ) {
                        $combined_tax_rate  += $rate->tax_rate;
                        $combined_tax_code  .= empty( $combined_tax_code ) ? WC_Tax::get_rate_code( $rate->tax_rate_id ) : '|' . WC_Tax::get_rate_code( $rate->tax_rate_id );
                        $combined_wc_tax_id .= empty( $combined_wc_tax_id ) ? $rate->tax_rate_id : '-' . $rate->tax_rate_id;
                    }

                    $combined_tax_rate = $combined_tax_rate / 100;
                    $tax_inclusive     = get_option( 'woocommerce_prices_include_tax' );
                    $tax_inclusive     = $tax_inclusive === "no" ? "false" : "true";

                    $result = $wpdb->get_var( $wpdb->prepare( "SELECT lightspeed_tax_category_id FROM $WCLSI_TAX_LOOKUP_TABLE WHERE woocommerce_tax_rate_id = %s", $combined_wc_tax_id ) );

                    $args = array(
                        $combined_wc_tax_id,
                        $combined_tax_code,
                        $combined_tax_rate,
                        $tax_inclusive,
                        $result
                    );

                    //schedule actions 2 seconds apart to avoid overloading the api
                    as_schedule_single_action( time() + 2 * $tax_categories_created ++, 'wclsi_process_tax_category', $args, 'wclsi' );

                }
            }


            echo json_encode( 'success' );
            die();
        }

        function process_combined_tax_ids( $combined_wc_tax_id, $combined_tax_code, $combined_tax_rate, $tax_inclusive, $previous_record ) {
            global $wpdb, $WCLSI_TAX_LOOKUP_TABLE;

            while ( wclsi_get_bucket_status() > 0.8 ) {
                sleep( 1 );
            }
            if ( is_null( $previous_record ) ) {
                $ls_tax_category_id = $this->create_or_update_ls_tax_category( $combined_tax_code, $combined_tax_rate, $tax_inclusive );
            } else {
                $ls_tax_category_id = $this->create_or_update_ls_tax_category( $combined_tax_code, $combined_tax_rate, $tax_inclusive, false, $previous_record );
            }

            if ( $ls_tax_category_id !== false ) {
                // Insert tax ids into lookup table or update if the woocommerce tax id is already present
                $sql = "INSERT INTO $WCLSI_TAX_LOOKUP_TABLE (lightspeed_category_name, woocommerce_tax_rate_id, lightspeed_tax_category_id) VALUES (%s, %s,%d) ON DUPLICATE KEY UPDATE lightspeed_tax_category_id = %d";
                $sql = $wpdb->prepare( $sql, $combined_tax_code, $combined_wc_tax_id, $ls_tax_category_id, $ls_tax_category_id );
                $wpdb->query( $sql );
            }

        }

        function get_ls_tax_id_for_wc_tax_rate( $wc_tax_rate_id ): ?int {
            global $wpdb, $WCLSI_TAX_LOOKUP_TABLE;

            $ls_tax_id = $wpdb->get_var( $wpdb->prepare( "SELECT lightspeed_tax_category_id FROM $WCLSI_TAX_LOOKUP_TABLE WHERE woocommerce_tax_rate_id = %s", $wc_tax_rate_id ) );

            if ( ! is_null( $ls_tax_id ) ) {
                return $ls_tax_id;
            } else {
                WCLSI_Logger::log_error( "No matching lightspeed tax category for woocommerce combined rate: $wc_tax_rate_id" );

                return null;
            }
        }

        function set_sales_tax_rate_id() {
            $tax_items    = isset( $this->order ) ? $this->order->get_items( 'tax' ) : $this->refund->get_items( 'tax' );
            $tax_rate_ids = [];
            foreach ( $tax_items as $item ) {
                $tax_rate_ids[] = $item->get_rate_id();
            }
            usort( $tax_rate_ids, function ( $a, $b ) {
                return ( (int) $a ) > ( (int) $b );
            } );

            $combined_tax_rate_id = implode( '-', $tax_rate_ids );

            //if there are tax rates, get the associated ls tax rate id
            if ( ! empty( $combined_tax_rate_id ) ) {
                $ls_tax_cat_id = $this->get_ls_tax_id_for_wc_tax_rate( $combined_tax_rate_id );
            }
            //this will have to change to allow multiple tax rates
            if ( ! empty( $ls_tax_cat_id ) ) {
                $this->ls_tax_id = $ls_tax_cat_id;
            } else {
                $this->ls_tax_id = null;
            }
        }

        function get_combined_ids( $base_array, $constructed_array ) {
            global $WCLSI_MAX_COMBINED_TAX_RATES;

            if ( ! count( $base_array ) > 0 ) {
                return $constructed_array;
            }

            if ( count( $constructed_array ) > 0 ) {
                $result = array();
                foreach ( $constructed_array as $element ) {
                    foreach ( $base_array[0] as $add_on ) {
                        if ( count( $element ) < $WCLSI_MAX_COMBINED_TAX_RATES ) {
                            $result[] = array_merge( $element, $add_on );
                        } else {
                            break;
                        }
                    }
                }
                $result = array_merge( $constructed_array, $result );
                array_shift( $base_array );
            } else {
                $result = array_shift( $base_array );
            }

            return $this->get_combined_ids( $base_array, $result );
        }


        function get_filtered_combined_ids( $ids_in_buckets ) {
            $result = $this->get_combined_ids( $ids_in_buckets, array() );

            $unique_combinations = array_unique( $result, SORT_REGULAR );
            $unique_combinations = array_filter( $unique_combinations, function ( $arr ) {
                return count( $arr ) > 0;
            } );
            $sorted_rates        = array();
            foreach ( $unique_combinations as $rate ) {
                if ( count( $rate ) > 0 ) {
                    usort( $rate, function ( $a, $b ) {
                        return ( (int) $a->tax_rate_id ) > ( (int) $b->tax_rate_id );
                    } );
                    $sorted_rates[] = $rate;
                }
            }

            return $sorted_rates;
        }

        function delete_ls_tax_category( $tax_category_id ) {
            global $WCLSI;

            $action       = 'Delete';
            $control_name = "Account.TaxCategory.$tax_category_id";


            $result = $WCLSI->make_api_call( $control_name, $action );

            if ( is_wp_error( $result ) ) {
                $msg = "Error deleting tax category";
                WCLSI_Logger::log_error( $msg, $result->get_error_message() );

                return false;
            }

            return true;
        }

        function delete_tax_categories_in_lookup_table_ajax() {
            wclsi_verify_nonce();
            global $wpdb, $WCLSI_TAX_LOOKUP_TABLE;

            $tax_category_ids = $wpdb->get_results( "SELECT lightspeed_tax_category_id FROM $WCLSI_TAX_LOOKUP_TABLE" );

            $tax_categories_deleted = 0;

            foreach ( $tax_category_ids as $tax_category_id ) {
                $id = $tax_category_id->lightspeed_tax_category_id;

                //schedule actions 2 seconds apart to avoid overloading the api
                as_schedule_single_action( time() + 2 * $tax_categories_deleted ++, 'wclsi_delete_tax_category', array( $id ), 'wclsi' );
            }

        }

        function delete_tax_category_and_remove_from_lookup_table( $id ) {
            global $wpdb, $WCLSI_TAX_LOOKUP_TABLE;

            while ( wclsi_get_bucket_status() > 0.8 ) {
                sleep( 1 );
            }

            $result = $this->delete_ls_tax_category( $id );
            if ( $result === true ) {
                $wpdb->delete( $WCLSI_TAX_LOOKUP_TABLE, array( 'lightspeed_tax_category_id' => $id ) );
            }
        }
    }

    global $WCLSI_SYNC_SALES;
    $WCLSI_SYNC_SALES = new LSI_Sync_Sales();
endif;